@@ -16,7 +16,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
-    $Id: alsa_driver.c,v 1.55 2004/09/13 02:11:46 trutkin Exp $
+    $Id: alsa_driver.c,v 1.68 2005/05/10 03:21:18 pbd Exp $
 */
 
 #include <math.h>
@@ -33,6 +33,7 @@
  
 #include <jack/internal.h>
 #include <jack/engine.h>
+#include <jack/messagebuffer.h>
 
 #include <sysdeps/time.h>
 
@@ -40,6 +41,7 @@
 #include "hammerfall.h"
 #include "hdsp.h"
 #include "ice1712.h"
+#include "usx2y.h"
 #include "generic.h"
 
 extern void store_work_time (int);
@@ -50,12 +52,14 @@ extern void show_work_times ();
 #undef DEBUG_WAKEUP
 
 /* Delay (in process calls) before jackd will report an xrun */
-#define XRUN_REPORT_DELAY 64
-
+#define XRUN_REPORT_DELAY 0
 
 static void
 alsa_driver_release_channel_dependent_memory (alsa_driver_t *driver)
 {
+	bitset_destroy (&driver->channels_done);
+	bitset_destroy (&driver->channels_not_done);
+
 	if (driver->playback_addr) {
 		free (driver->playback_addr);
 		driver->playback_addr = 0;
@@ -152,6 +156,13 @@ alsa_driver_ice1712_hardware (alsa_driver_t *driver)
         return 0;
 }
 
+static int
+alsa_driver_usx2y_hardware (alsa_driver_t *driver)
+{
+	driver->hw = jack_alsa_usx2y_hw_new (driver);
+	return 0;
+}
+
 static int
 alsa_driver_generic_hardware (alsa_driver_t *driver)
 {
@@ -177,6 +188,10 @@ alsa_driver_hw_specific (alsa_driver_t *driver, int hw_monitoring,
                 if ((err = alsa_driver_ice1712_hardware (driver)) !=0) {
                         return err;
                 }
+	} else if (!strcmp(driver->alsa_driver, "USB US-X2Y")) {
+		if ((err = alsa_driver_usx2y_hardware (driver)) !=0) {
+				return err;
+		}
 	} else {
 	        if ((err = alsa_driver_generic_hardware (driver)) != 0) {
 			return err;
@@ -264,7 +279,8 @@ alsa_driver_setup_io_function_pointers (alsa_driver_t *driver)
 
 			case Shaped:
 			printf("Noise-shaped dithering at 16 bits\n");
-			driver->write_via_copy = sample_move_dither_shaped_d24_sS;
+			driver->write_via_copy =
+				sample_move_dither_shaped_d24_sS;
 			break;
 
 			default:
@@ -325,9 +341,12 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 			      snd_pcm_t *handle, 
 			      snd_pcm_hw_params_t *hw_params, 
 			      snd_pcm_sw_params_t *sw_params, 
-			      unsigned long *nchns,unsigned long sample_width)
+			      unsigned int *nperiodsp,
+			      unsigned long *nchns,
+			      unsigned long sample_width)
 {
 	int err, format;
+	unsigned int frame_rate;
 	snd_pcm_uframes_t stop_th;
 	static struct {
 		char Name[16];
@@ -338,7 +357,7 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 		{"16bit", SND_PCM_FORMAT_S16},
 	};
 #define NOFORMATS (sizeof(formats)/sizeof(formats[0]))
-	
+
 	if ((err = snd_pcm_hw_params_any (handle, hw_params)) < 0)  {
 		jack_error ("ALSA: no playback configurations available (%s)",
 			    snd_strerror (err));
@@ -368,11 +387,16 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 
 	format = sample_width == 4 ? 0 : NOFORMATS - 1;
 	while (1) {
-		if ((err = snd_pcm_hw_params_set_format (handle, hw_params, formats[format].format)) < 0) {
+		if ((err = snd_pcm_hw_params_set_format (
+			     handle, hw_params, formats[format].format)) < 0) {
 			int failed_format = format; 
-			if (sample_width == 4 ? format++ < NOFORMATS - 1 : format-- > 0) {
-				jack_error ("Couldn't open %s for %s samples"
-					    " trying %s instead", device_name, formats[failed_format].Name, formats[format].Name);
+			if ((sample_width == 4
+			     ? format++ < NOFORMATS - 1
+			     : format-- > 0)) {
+				jack_error ("Note: audio device %s doesn't support a %s sample format"
+					    " so JACK will try a %s format instead", device_name,
+					    formats[failed_format].Name,
+					    formats[format].Name);
 			} else {
 				jack_error ("Sorry. The audio interface \"%s\""
 					    " doesn't support any of the"
@@ -385,9 +409,11 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 			break;
 	} 
 
-	if ((err = snd_pcm_hw_params_set_rate_near (handle, hw_params,
-						    driver->frame_rate, 0))
-	    < 0) {
+	frame_rate = driver->frame_rate ;
+	err = snd_pcm_hw_params_set_rate_near (handle, hw_params,
+					       &frame_rate, NULL) ;
+	driver->frame_rate = frame_rate ;
+	if (err < 0) {
 		jack_error ("ALSA: cannot set sample/frame rate to %"
 			    PRIu32 " for %s", driver->frame_rate,
 			    stream_name);
@@ -396,7 +422,10 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 	if (!*nchns) {
 		/*if not user-specified, try to find the maximum
 		 * number of channels */
-		*nchns = snd_pcm_hw_params_get_channels_max (hw_params);
+		unsigned int channels_max ;
+		err = snd_pcm_hw_params_get_channels_max (hw_params,
+							  &channels_max);
+		*nchns = channels_max ;
 
 		if (*nchns > 1024) { 
 
@@ -436,13 +465,24 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 		return -1;
 	}
 
-	if ((err = snd_pcm_hw_params_set_periods (handle, hw_params,
-						  driver->user_nperiods, 0))
-	    < 0) {
+	*nperiodsp = driver->user_nperiods;
+	snd_pcm_hw_params_set_periods_min (handle, hw_params, nperiodsp, NULL);
+	if (*nperiodsp < driver->user_nperiods)
+		*nperiodsp = driver->user_nperiods;
+	if (snd_pcm_hw_params_set_periods_near (handle, hw_params,
+						nperiodsp, NULL) < 0) {
 		jack_error ("ALSA: cannot set number of periods to %u for %s",
-			    driver->user_nperiods, stream_name);
+			    *nperiodsp, stream_name);
 		return -1;
 	}
+
+	if (*nperiodsp < driver->user_nperiods) {
+		jack_error ("ALSA: got smaller periods %u than %u for %s",
+			    *nperiodsp, (unsigned int) driver->user_nperiods,
+			    stream_name);
+		return -1;
+	}
+	fprintf(stderr, "nperiods = %d for %s\n", *nperiodsp, stream_name);
 	
 	if (!jack_power_of_two(driver->frames_per_cycle)) {
 		jack_error("JACK: frames must be a power of two "
@@ -451,12 +491,12 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 	}
 
 	if ((err = snd_pcm_hw_params_set_buffer_size (handle, hw_params,
-						      driver->user_nperiods *
+						      *nperiodsp *
 						      driver->frames_per_cycle))
 	    < 0) {
 		jack_error ("ALSA: cannot set buffer length to %" PRIu32
 			    " for %s",
-			    driver->user_nperiods * driver->frames_per_cycle,
+			    *nperiodsp * driver->frames_per_cycle,
 			    stream_name);
 		return -1;
 	}
@@ -475,7 +515,7 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 		return -1;
 	}
 
-	stop_th = driver->user_nperiods * driver->frames_per_cycle;
+	stop_th = *nperiodsp * driver->frames_per_cycle;
 	if (driver->soft_mode) {
 		stop_th = (snd_pcm_uframes_t)-1;
 	}
@@ -496,20 +536,28 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 
 #if 0
 	fprintf (stderr, "set silence size to %lu * %lu = %lu\n",
-		 driver->frames_per_cycle, driver->user_nperiods,
-		 driver->frames_per_cycle * driver->user_nperiods);
+		 driver->frames_per_cycle, *nperiodsp,
+		 driver->frames_per_cycle * *nperiodsp);
 
 	if ((err = snd_pcm_sw_params_set_silence_size (
 		     handle, sw_params,
-		     driver->frames_per_cycle * driver->user_nperiods)) < 0) {
+		     driver->frames_per_cycle * *nperiodsp)) < 0) {
 		jack_error ("ALSA: cannot set silence size for %s",
 			    stream_name);
 		return -1;
 	}
 #endif
 
-	if ((err = snd_pcm_sw_params_set_avail_min (
-		     handle, sw_params, driver->frames_per_cycle)) < 0) {
+	if (handle == driver->playback_handle)
+		err = snd_pcm_sw_params_set_avail_min (
+			handle, sw_params,
+			driver->frames_per_cycle
+			* (*nperiodsp - driver->user_nperiods + 1));
+	else
+		err = snd_pcm_sw_params_set_avail_min (
+			handle, sw_params, driver->frames_per_cycle);
+			
+	if (err < 0) {
 		jack_error ("ALSA: cannot set avail min for %s", stream_name);
 		return -1;
 	}
@@ -530,13 +578,12 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 			    jack_nframes_t rate)
 {
 	int dir;
-	unsigned int p_period_size = 0;
-	unsigned int c_period_size = 0;
-	unsigned int p_nfragments = 0;
-	unsigned int c_nfragments = 0;
+	snd_pcm_uframes_t p_period_size = 0;
+	snd_pcm_uframes_t c_period_size = 0;
 	channel_t chn;
 	unsigned int pr = 0;
 	unsigned int cr = 0;
+	int err;
 
 	driver->frame_rate = rate;
 	driver->frames_per_cycle = frames_per_cycle;
@@ -554,6 +601,7 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 			    driver->capture_handle,
 			    driver->capture_hw_params,
 			    driver->capture_sw_params,
+			    &driver->capture_nperiods,
 			    &driver->capture_nchannels,
 			    driver->capture_sample_bytes)) {
 			jack_error ("ALSA: cannot configure capture channel");
@@ -569,6 +617,7 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 			    driver->playback_handle,
 			    driver->playback_hw_params,
 			    driver->playback_sw_params,
+			    &driver->playback_nperiods,
 			    &driver->playback_nchannels,
 			    driver->playback_sample_bytes)) {
 			jack_error ("ALSA: cannot configure playback channel");
@@ -579,39 +628,43 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 	/* check the rate, since thats rather important */
 
 	if (driver->playback_handle) {
-		int dir;
-		pr = snd_pcm_hw_params_get_rate (driver->playback_hw_params, &dir);
+		snd_pcm_hw_params_get_rate (driver->playback_hw_params,
+					    &pr, &dir);
 	}
 
 	if (driver->capture_handle) {
-		int dir;
-		cr = snd_pcm_hw_params_get_rate (driver->capture_hw_params, &dir);
+		snd_pcm_hw_params_get_rate (driver->capture_hw_params,
+					    &cr, &dir);
 	}
 
 	if (driver->capture_handle && driver->playback_handle) {
 		if (cr != pr) {
-			jack_error ("playback and capture sample rates do not match (%d vs. %d)",
-				    pr, cr);
+			jack_error ("playback and capture sample rates do "
+				    "not match (%d vs. %d)", pr, cr);
 		}
 
-		/* only change if *both* capture and playback rates don't match requested
-		 * certain hardware actually still works properly in full-duplex with
-		 * slightly different rate values between adc and dac
+		/* only change if *both* capture and playback rates
+		 * don't match requested certain hardware actually
+		 * still works properly in full-duplex with slightly
+		 * different rate values between adc and dac
 		 */
 		if (cr != driver->frame_rate && pr != driver->frame_rate) {
-			jack_error ("sample rate in use (%d Hz) does not match requested rate (%d Hz)",
+			jack_error ("sample rate in use (%d Hz) does not "
+				    "match requested rate (%d Hz)",
 				    cr, driver->frame_rate);
 			driver->frame_rate = cr;
 		}
 		
 	}
 	else if (driver->capture_handle && cr != driver->frame_rate) {
-		jack_error ("capture sample rate in use (%d Hz) does not match requested rate (%d Hz)",
+		jack_error ("capture sample rate in use (%d Hz) does not "
+			    "match requested rate (%d Hz)",
 			    cr, driver->frame_rate);
 		driver->frame_rate = cr;
 	}
 	else if (driver->playback_handle && pr != driver->frame_rate) {
-		jack_error ("playback sample rate in use (%d Hz) does not match requested rate (%d Hz)",
+		jack_error ("playback sample rate in use (%d Hz) does not "
+			    "match requested rate (%d Hz)",
 			    pr, driver->frame_rate);
 		driver->frame_rate = pr;
 	}
@@ -620,19 +673,17 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 	/* check the fragment size, since thats non-negotiable */
 	
 	if (driver->playback_handle) {
-		p_period_size =
-			snd_pcm_hw_params_get_period_size (
-				driver->playback_hw_params, &dir);
-		p_nfragments =
-			snd_pcm_hw_params_get_periods (
-				driver->playback_hw_params, &dir);
-		driver->playback_sample_format = (snd_pcm_format_t)
-			snd_pcm_hw_params_get_format (
-				driver->playback_hw_params);
-		driver->playback_interleaved =
-			(snd_pcm_hw_params_get_access (
-				driver->playback_hw_params)
-			 == SND_PCM_ACCESS_MMAP_INTERLEAVED);
+ 		snd_pcm_access_t access;
+
+ 		err = snd_pcm_hw_params_get_period_size (
+ 			driver->playback_hw_params, &p_period_size, &dir);
+ 		err = snd_pcm_hw_params_get_format (
+ 			driver->playback_hw_params,
+			&(driver->playback_sample_format));
+ 		err = snd_pcm_hw_params_get_access (driver->playback_hw_params,
+						    &access);
+ 		driver->playback_interleaved =
+			(access == SND_PCM_ACCESS_MMAP_INTERLEAVED);
 
 		if (p_period_size != driver->frames_per_cycle) {
 			jack_error ("alsa_pcm: requested an interrupt every %"
@@ -644,19 +695,18 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 	}
 
 	if (driver->capture_handle) {
-		c_period_size =
-			snd_pcm_hw_params_get_period_size (
-				driver->capture_hw_params, &dir);
-		c_nfragments =
-			snd_pcm_hw_params_get_periods (
-				driver->capture_hw_params, &dir);
-		driver->capture_sample_format = (snd_pcm_format_t)
-			snd_pcm_hw_params_get_format (
-				driver->capture_hw_params);
-		driver->capture_interleaved =
-			(snd_pcm_hw_params_get_access (
-				driver->capture_hw_params)
-			 == SND_PCM_ACCESS_MMAP_INTERLEAVED);
+ 		snd_pcm_access_t access;
+
+ 		err = snd_pcm_hw_params_get_period_size (
+ 			driver->capture_hw_params, &c_period_size, &dir);
+ 		err = snd_pcm_hw_params_get_format (
+ 			driver->capture_hw_params,
+			&(driver->capture_sample_format));
+ 		err = snd_pcm_hw_params_get_access (driver->capture_hw_params,
+						    &access);
+ 		driver->capture_interleaved =
+			(access == SND_PCM_ACCESS_MMAP_INTERLEAVED);
+ 
 	
 		if (c_period_size != driver->frames_per_cycle) {
 			jack_error ("alsa_pcm: requested an interrupt every %"
@@ -667,26 +717,6 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 		}
 	}
 
-	if (driver->capture_handle && driver->playback_handle) {
-		if (p_nfragments != c_nfragments) {
-			jack_error ("alsa_pcm: different period counts for "
-				    "playback and capture!");
-			return -1;
-		}
-
-		driver->nfragments = c_nfragments;
-
-	} else if (driver->capture_handle) {
-
-		driver->nfragments = c_nfragments;
-
-	} else {
-
-		driver->nfragments = p_nfragments;
-
-	}
-
-	driver->buffer_frames = driver->frames_per_cycle * driver->nfragments;
 	driver->playback_sample_bytes =
 		snd_pcm_format_physical_width (driver->playback_sample_format)
 		/ 8;
@@ -779,12 +809,10 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 	   channels require action on every cycle. any bits that are
 	   not set after the engine's process() call indicate channels
 	   that potentially need to be silenced.
-
-	   XXX this is limited to <wordsize> channels. Use a bitset
-	   type instead.
 	*/
 
-	driver->channel_done_bits = 0;
+	bitset_create (&driver->channels_done, driver->max_nchannels);
+	bitset_create (&driver->channels_not_done, driver->max_nchannels);
 
 	if (driver->playback_handle) {
 		driver->playback_addr = (char **)
@@ -800,7 +828,7 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 		}
 
 		for (chn = 0; chn < driver->playback_nchannels; chn++) {
-			driver->channel_done_bits |= (1<<chn);
+			bitset_add (driver->channels_done, chn);
 		}
 
 		driver->dither_state = (dither_state_t *)
@@ -923,67 +951,16 @@ alsa_driver_start (alsa_driver_t *driver)
 	}
 
 	if (driver->hw_monitoring) {
-		driver->hw->set_input_monitor_mask (driver->hw,
-						    driver->input_monitor_mask);
-	}
-
-	if (driver->playback_handle) {
-		/* fill playback buffer with zeroes, and mark 
-		   all fragments as having data.
-		*/
-		
-		pavail = snd_pcm_avail_update (driver->playback_handle);
-
-		if (pavail != driver->buffer_frames) {
-			jack_error ("ALSA: full buffer not available at start");
-			return -1;
-		}
-	
-		if (alsa_driver_get_channel_addresses (driver, 0, &pavail,
-						       0, &poffset)) {
-			return -1;
-		}
-
-		/* XXX this is cheating. ALSA offers no guarantee that
-		   we can access the entire buffer at any one time. It
-		   works on most hardware tested so far, however, buts
-		   its a liability in the long run. I think that
-		   alsa-lib may have a better function for doing this
-		   here, where the goal is to silence the entire
-		   buffer.
-		*/
-		
-		for (chn = 0; chn < driver->playback_nchannels; chn++) {
-			alsa_driver_silence_on_channel (driver, chn,
-							driver->buffer_frames);
-		}
-		
-		snd_pcm_mmap_commit (driver->playback_handle, poffset,
-				     driver->buffer_frames);
-		
-		if ((err = snd_pcm_start (driver->playback_handle)) < 0) {
-			jack_error ("could not start playback (%s)",
-				    snd_strerror (err));
-			return -1;
-		}
-	}
-
-	if ((driver->capture_handle && driver->capture_and_playback_not_synced)
-	    || !driver->playback_handle) {
-		if ((err = snd_pcm_start (driver->capture_handle)) < 0) {
-			jack_error ("could not start capture (%s)",
-				    snd_strerror (err));
-			return -1;
-		}
-	}
-			
-	if (driver->hw_monitoring &&
-	    (driver->input_monitor_mask || driver->all_monitor_in)) {
-		if (driver->all_monitor_in) {
-			driver->hw->set_input_monitor_mask (driver->hw, ~0U);
+		if (driver->input_monitor_mask || driver->all_monitor_in) {
+			if (driver->all_monitor_in) {
+				driver->hw->set_input_monitor_mask (driver->hw, ~0U);
+			} else {
+				driver->hw->set_input_monitor_mask (
+					driver->hw, driver->input_monitor_mask);
+			}
 		} else {
-			driver->hw->set_input_monitor_mask (
-				driver->hw, driver->input_monitor_mask);
+			driver->hw->set_input_monitor_mask (driver->hw,
+							    driver->input_monitor_mask);
 		}
 	}
 
@@ -1009,6 +986,60 @@ alsa_driver_start (alsa_driver_t *driver)
 		malloc (sizeof (struct pollfd) * 
 			(driver->playback_nfds + driver->capture_nfds + 2));
 
+	if (driver->playback_handle) {
+		/* fill playback buffer with zeroes, and mark 
+		   all fragments as having data.
+		*/
+		
+		pavail = snd_pcm_avail_update (driver->playback_handle);
+
+		if (pavail !=
+		    driver->frames_per_cycle * driver->playback_nperiods) {
+			jack_error ("ALSA: full buffer not available at start");
+			return -1;
+		}
+	
+		if (alsa_driver_get_channel_addresses (driver,
+					0, &pavail, 0, &poffset)) {
+			return -1;
+		}
+
+		/* XXX this is cheating. ALSA offers no guarantee that
+		   we can access the entire buffer at any one time. It
+		   works on most hardware tested so far, however, buts
+		   its a liability in the long run. I think that
+		   alsa-lib may have a better function for doing this
+		   here, where the goal is to silence the entire
+		   buffer.
+		*/
+		
+		for (chn = 0; chn < driver->playback_nchannels; chn++) {
+			alsa_driver_silence_on_channel (
+				driver, chn,
+				driver->user_nperiods
+				* driver->frames_per_cycle);
+		}
+		
+		snd_pcm_mmap_commit (driver->playback_handle, poffset,
+				     driver->user_nperiods
+				     * driver->frames_per_cycle);
+		
+		if ((err = snd_pcm_start (driver->playback_handle)) < 0) {
+			jack_error ("ALSA: could not start playback (%s)",
+				    snd_strerror (err));
+			return -1;
+		}
+	}
+
+	if ((driver->capture_handle && driver->capture_and_playback_not_synced)
+	    || !driver->playback_handle) {
+		if ((err = snd_pcm_start (driver->capture_handle)) < 0) {
+			jack_error ("ALSA: could not start capture (%s)",
+				    snd_strerror (err));
+			return -1;
+		}
+	}
+			
 	return 0;
 }
 
@@ -1037,7 +1068,7 @@ alsa_driver_stop (alsa_driver_t *driver)
 		
 	if (driver->playback_handle) {
 		if ((err = snd_pcm_drop (driver->playback_handle)) < 0) {
-			jack_error ("alsa_pcm: channel flush for playback "
+			jack_error ("ALSA: channel flush for playback "
 				    "failed (%s)", snd_strerror (err));
 			return -1;
 		}
@@ -1047,7 +1078,7 @@ alsa_driver_stop (alsa_driver_t *driver)
 	    || driver->capture_and_playback_not_synced) {
 		if (driver->capture_handle) {
 			if ((err = snd_pcm_drop (driver->capture_handle)) < 0) {
-				jack_error ("alsa_pcm: channel flush for "
+				jack_error ("ALSA: channel flush for "
 					    "capture failed (%s)",
 					    snd_strerror (err));
 				return -1;
@@ -1063,7 +1094,15 @@ alsa_driver_stop (alsa_driver_t *driver)
 }
 
 static int
-alsa_driver_xrun_recovery (alsa_driver_t *driver)
+alsa_driver_restart (alsa_driver_t *driver)
+{
+ 	if (driver->nt_stop((struct _jack_driver_nt *) driver))
+ 	    return -1;
+ 	return driver->nt_start((struct _jack_driver_nt *) driver);
+}
+
+static int
+alsa_driver_xrun_recovery (alsa_driver_t *driver, float *delayed_usecs)
 {
 	snd_pcm_status_t *status;
 	int res;
@@ -1089,28 +1128,29 @@ alsa_driver_xrun_recovery (alsa_driver_t *driver)
 		gettimeofday(&now, 0);
 		snd_pcm_status_get_trigger_tstamp(status, &tstamp);
 		timersub(&now, &tstamp, &diff);
-		fprintf(stderr, "\n\n**** alsa_pcm: xrun of at least %.3f "
+		*delayed_usecs = diff.tv_sec * 1000000.0 + diff.tv_usec;
+		MESSAGE("\n\n**** alsa_pcm: xrun of at least %.3f "
 			"msecs\n\n",
-			diff.tv_sec * 1000 + diff.tv_usec / 1000.0);
+			*delayed_usecs / 1000.0);
 	}
 
-	if (alsa_driver_stop (driver) ||
-	    alsa_driver_start (driver)) {
+	if (alsa_driver_restart (driver)) {
 		return -1;
-
 	}
 	return 0;
 }	
 
-static void
+void
 alsa_driver_silence_untouched_channels (alsa_driver_t *driver,
 					jack_nframes_t nframes)
 {
 	channel_t chn;
+	jack_nframes_t buffer_frames =
+		driver->frames_per_cycle * driver->playback_nperiods;
 
 	for (chn = 0; chn < driver->playback_nchannels; chn++) {
-		if ((driver->channels_not_done & (1<<chn))) { 
-			if (driver->silent[chn] < driver->buffer_frames) {
+		if (bitset_contains (driver->channels_not_done, chn)) { 
+			if (driver->silent[chn] < buffer_frames) {
 				alsa_driver_silence_on_channel_no_mark (
 					driver, chn, nframes);
 				driver->silent[chn] += nframes;
@@ -1194,6 +1234,16 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 
 		poll_enter = jack_get_microseconds ();
 
+		if (poll_enter > driver->poll_next) {
+			/*
+			 * This processing cycle was delayed past the
+			 * next due interrupt!  Do not account this as
+			 * a wakeup delay:
+			 */
+			driver->poll_next = 0;
+			driver->poll_late++;
+		}
+
 		if (poll (driver->pfd, nfds, driver->poll_timeout) < 0) {
 
 			if (errno == EINTR) {
@@ -1348,7 +1398,7 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 	}
 
 	if (xrun_detected) {
-		*status = alsa_driver_xrun_recovery (driver);
+		*status = alsa_driver_xrun_recovery (driver, delayed_usecs);
 		return 0;
 	}
 
@@ -1365,7 +1415,7 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 
 	/* mark all channels not done for now. read/write will change this */
 
-	driver->channels_not_done = driver->channel_done_bits;
+	bitset_copy (driver->channels_not_done, driver->channels_done);
 
 	/* constrain the available count to the nearest (round down) number of
 	   periods.
@@ -1382,13 +1432,15 @@ alsa_driver_null_cycle (alsa_driver_t* driver, jack_nframes_t nframes)
 	snd_pcm_uframes_t contiguous;
 	int chn;
 
+	if (nframes > driver->frames_per_cycle) {
+		return -1;
+	}
+	
 	if (driver->capture_handle) {
 		nf = nframes;
 		offset = 0;
 		while (nf) {
-			
-			contiguous = (nf > driver->frames_per_cycle) ?
-				driver->frames_per_cycle : nf;
+			contiguous = nf;
 			
 			if (snd_pcm_mmap_begin (
 				    driver->capture_handle,
@@ -1411,8 +1463,7 @@ alsa_driver_null_cycle (alsa_driver_t* driver, jack_nframes_t nframes)
 		nf = nframes;
 		offset = 0;
 		while (nf) {
-			contiguous = (nf > driver->frames_per_cycle) ?
-				driver->frames_per_cycle : nf;
+			contiguous = nf;
 			
 			if (snd_pcm_mmap_begin (
 				    driver->playback_handle,
@@ -1453,22 +1504,27 @@ alsa_driver_read (alsa_driver_t *driver, jack_nframes_t nframes)
 	snd_pcm_sframes_t contiguous;
 	snd_pcm_sframes_t nread;
 	snd_pcm_sframes_t offset;
+	jack_nframes_t  orig_nframes;
 	jack_default_audio_sample_t* buf;
 	channel_t chn;
 	JSList *node;
 	jack_port_t* port;
+	int err;
 
 	if (!driver->capture_handle || driver->engine->freewheeling) {
 		return 0;
 	}
-
+	if (nframes > driver->frames_per_cycle) {
+		return -1;
+	}
+	
 	nread = 0;
 	contiguous = 0;
-
+	orig_nframes = nframes;
+	
 	while (nframes) {
 		
-		contiguous = (nframes > driver->frames_per_cycle) ?
-			driver->frames_per_cycle : nframes;
+		contiguous = nframes;
 		
 		if (alsa_driver_get_channel_addresses (
 			    driver, 
@@ -1487,21 +1543,20 @@ alsa_driver_read (alsa_driver_t *driver, jack_nframes_t nframes)
 				/* no-copy optimization */
 				continue;
 			}
-
-			buf = jack_port_get_buffer (port, nframes);
+			buf = jack_port_get_buffer (port, orig_nframes);
 			alsa_driver_read_from_channel (driver, chn,
-						       buf + nread,
-						       contiguous);
+				buf + nread, contiguous);
 		}
 		
-		if (snd_pcm_mmap_commit (driver->capture_handle, offset,
-					 contiguous) < 0) {
-			jack_error ("alsa_pcm: could not complete read of %"
-				    PRIu32 " frames\n", contiguous);
+		if ((err = snd_pcm_mmap_commit (driver->capture_handle,
+				offset, contiguous)) < 0) {
+			jack_error ("ALSA: could not complete read of %"
+				PRIu32 " frames: error = %d\n", contiguous, err);
 			return -1;
 		}
 
 		nframes -= contiguous;
+		nread += contiguous;
 	}
 
 	return 0;
@@ -1512,7 +1567,10 @@ alsa_driver_write (alsa_driver_t* driver, jack_nframes_t nframes)
 {
 	channel_t chn;
 	JSList *node;
+	JSList *mon_node;
 	jack_default_audio_sample_t* buf;
+	jack_default_audio_sample_t* monbuf;
+	jack_nframes_t orig_nframes;
 	snd_pcm_sframes_t nwritten;
 	snd_pcm_sframes_t contiguous;
 	snd_pcm_sframes_t offset;
@@ -1524,9 +1582,13 @@ alsa_driver_write (alsa_driver_t* driver, jack_nframes_t nframes)
 	if (!driver->playback_handle || driver->engine->freewheeling) {
 		return 0;
 	}
+	if (nframes > driver->frames_per_cycle) {
+		return -1;
+	}
 
 	nwritten = 0;
 	contiguous = 0;
+	orig_nframes = nframes;
 	
 	/* check current input monitor request status */
 	
@@ -1550,8 +1612,7 @@ alsa_driver_write (alsa_driver_t* driver, jack_nframes_t nframes)
 	
 	while (nframes) {
 		
-		contiguous = (nframes > driver->frames_per_cycle) ?
-			driver->frames_per_cycle : nframes;
+		contiguous = nframes;
 		
 		if (alsa_driver_get_channel_addresses (
 			    driver, 
@@ -1561,7 +1622,8 @@ alsa_driver_write (alsa_driver_t* driver, jack_nframes_t nframes)
 			return -1;
 		}
 		
-		for (chn = 0, node = driver->playback_ports; node;
+		for (chn = 0, node = driver->playback_ports, mon_node=driver->monitor_ports;
+		     node;
 		     node = jack_slist_next (node), chn++) {
 
 			port = (jack_port_t *) node->data;
@@ -1569,30 +1631,37 @@ alsa_driver_write (alsa_driver_t* driver, jack_nframes_t nframes)
 			if (!jack_port_connected (port)) {
 				continue;
 			}
-			
-			buf = jack_port_get_buffer (port, contiguous);
-
+			buf = jack_port_get_buffer (port, orig_nframes);
 			alsa_driver_write_to_channel (driver, chn,
-						      buf + nwritten,
-						      contiguous);
+				buf + nwritten, contiguous);
+
+			if (mon_node) {
+				port = (jack_port_t *) mon_node->data;
+				if (!jack_port_connected (port)) {
+					continue;
+				}
+				monbuf = jack_port_get_buffer (port, orig_nframes);
+				memcpy (monbuf + nwritten, buf + nwritten, contiguous * sizeof(jack_default_audio_sample_t));
+				mon_node = jack_slist_next (mon_node);				
+			}
 		}
 
-		if (driver->channels_not_done) {
+		
+		if (!bitset_empty (driver->channels_not_done)) {
 			alsa_driver_silence_untouched_channels (driver,
 								contiguous);
 		}
 		
 		if ((err = snd_pcm_mmap_commit (driver->playback_handle,
-						offset, contiguous))
-		    < 0) {
-			jack_error ("could not complete playback of %"
-				    PRIu32 " frames: error = %d",
-				    contiguous, err);
+				offset, contiguous)) < 0) {
+			jack_error ("ALSA: could not complete playback of %"
+				PRIu32 " frames: error = %d", contiguous, err);
 			if (err != EPIPE && err != ESTRPIPE)
 				return -1;
 		}
 
 		nframes -= contiguous;
+		nwritten += contiguous;
 	}
 
 	return 0;
@@ -1610,15 +1679,14 @@ alsa_driver_run_cycle (alsa_driver_t *driver)
 
 	nframes = alsa_driver_wait (driver, -1, &wait_status, &delayed_usecs);
 
-	DEBUG ("alsa back from wait, nframes = %lu", nframes);
+	DEBUG ("alsaback from wait, nframes = %lu", nframes);
 
 	if (nframes == 0) {
 
 		/* we detected an xrun and restarted: notify
 		 * clients about the delay. 
 		 */
-
-		engine->delay (engine);
+		engine->delay (engine, delayed_usecs);
 		return 0;
 	} 
 
@@ -1659,10 +1727,7 @@ alsa_driver_attach (alsa_driver_t *driver)
 			break;
 		}
 
-		/* XXX fix this so that it can handle: systemic
-		 * (external) latency
-		*/
-		jack_port_set_latency (port, driver->frames_per_cycle);
+		jack_port_set_latency (port, driver->frames_per_cycle + driver->capture_frame_latency);
 
 		driver->capture_ports =
 			jack_slist_append (driver->capture_ports, port);
@@ -1682,11 +1747,7 @@ alsa_driver_attach (alsa_driver_t *driver)
 			break;
 		}
 		
-		/* XXX fix this so that it can handle: systemic
-		 * (external) latency
-		*/
-		jack_port_set_latency (port, driver->frames_per_cycle
-				       * driver->nfragments);
+		jack_port_set_latency (port, (driver->frames_per_cycle * (driver->user_nperiods - 1)) + driver->playback_frame_latency);
 
 		driver->playback_ports =
 			jack_slist_append (driver->playback_ports, port);
@@ -1701,8 +1762,13 @@ alsa_driver_attach (alsa_driver_t *driver)
 				jack_error ("ALSA: cannot register monitor "
 					    "port for %s", buf);
 			} else {
-				jack_port_tie (port, monitor_port);
+
+				jack_port_set_latency (monitor_port, driver->frames_per_cycle);
+				
+				driver->monitor_ports =
+					jack_slist_append (driver->monitor_ports, monitor_port);
 			}
+			
 		}
 	}
 
@@ -1737,6 +1803,17 @@ alsa_driver_detach (alsa_driver_t *driver)
 	jack_slist_free (driver->playback_ports);
 	driver->playback_ports = 0;
 
+	if (driver->monitor_ports) {
+		for (node = driver->monitor_ports; node;
+		     node = jack_slist_next (node)) {
+			jack_port_unregister (driver->client,
+					      ((jack_port_t *) node->data));
+		}
+		
+		jack_slist_free (driver->monitor_ports);
+		driver->monitor_ports = 0;
+	}
+	
 	return 0;
 }
 
@@ -1863,7 +1940,9 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 		 int monitor,
 		 int user_capture_nchnls,
 		 int user_playback_nchnls,
-		 int shorts_first
+		 int shorts_first,
+		 jack_nframes_t capture_latency,
+		 jack_nframes_t playback_latency
 		 )
 {
 	int err;
@@ -1901,13 +1980,14 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 	driver->ctl_handle = 0;
 	driver->hw = 0;
 	driver->capture_and_playback_not_synced = FALSE;
-	driver->nfragments = 0;
 	driver->max_nchannels = 0;
 	driver->user_nchannels = 0;
 	driver->playback_nchannels = user_playback_nchnls;
 	driver->capture_nchannels = user_capture_nchnls;
 	driver->playback_sample_bytes = (shorts_first ? 2:4);
 	driver->capture_sample_bytes = (shorts_first ? 2:4);
+	driver->capture_frame_latency = capture_latency;
+	driver->playback_frame_latency = playback_latency;
 
 	driver->playback_addr = 0;
 	driver->capture_addr = 0;
@@ -1922,7 +2002,8 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 
 	driver->capture_ports = 0;
 	driver->playback_ports = 0;
-
+	driver->monitor_ports = 0;
+	
 	driver->pfd = 0;
 	driver->playback_nfds = 0;
 	driver->capture_nfds = 0;
@@ -1933,9 +2014,20 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 	pthread_mutex_init (&driver->clock_sync_lock, 0);
 	driver->clock_sync_listeners = 0;
 
+	driver->poll_late = 0;
 	driver->xrun_count = 0;
 	driver->process_count = 0;
 
+	driver->alsa_name_playback = strdup (playback_alsa_device);
+	driver->alsa_name_capture = strdup (capture_alsa_device);
+
+	if (alsa_driver_check_card_type (driver)) {
+		alsa_driver_delete (driver);
+		return NULL;
+	}
+
+	alsa_driver_hw_specific (driver, hw_monitoring, hw_metering);
+
 	if (playing) {
 		if (snd_pcm_open (&driver->playback_handle,
 				  playback_alsa_device,
@@ -1948,6 +2040,7 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 					    " application using it and "
 					    "run JACK again",
 					    playback_alsa_device);
+				alsa_driver_delete (driver);
 				return NULL;
 				break;
 
@@ -1955,6 +2048,7 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 				jack_error ("you do not have permission to open "
 					    "the audio device \"%s\" for playback",
 					    playback_alsa_device);
+				alsa_driver_delete (driver);
 				return NULL;
 				break;
 			} 
@@ -1979,6 +2073,7 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 					    " application using it and "
 					    "run JACK again",
 					    capture_alsa_device);
+				alsa_driver_delete (driver);
 				return NULL;
 				break;
 
@@ -1986,6 +2081,7 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 				jack_error ("you do not have permission to open "
 					    "the audio device \"%s\" for capture",
 					    capture_alsa_device);
+				alsa_driver_delete (driver);
 				return NULL;
 				break;
 			} 
@@ -2009,8 +2105,8 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 
 			if (driver->capture_handle == NULL) {
 				/* can't do anything */
-				free (driver);
-				return 0;
+				alsa_driver_delete (driver);
+				return NULL;
 			}
 			
 			playing = FALSE;
@@ -2028,28 +2124,14 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 			
 			if (driver->playback_handle == NULL) {
 				/* can't do anything */
-				free (driver);
-				return 0;
+				alsa_driver_delete (driver);
+				return NULL;
 			}
 
 			capturing = FALSE;
 		}
 	}
 
-	driver->alsa_name_playback = strdup (playback_alsa_device);
-	driver->alsa_name_capture = strdup (capture_alsa_device);
-
-	if (alsa_driver_check_card_type (driver)) {
-		if (driver->capture_handle) {
-			snd_pcm_close (driver->capture_handle);
-		}
-		if (driver->playback_handle) {
-			snd_pcm_close (driver->playback_handle);
-		}
-		free (driver);
-		return 0;
-	}
-
 	driver->playback_hw_params = 0;
 	driver->capture_hw_params = 0;
 	driver->playback_sw_params = 0;
@@ -2061,7 +2143,7 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 			jack_error ("ALSA: could not allocate playback hw"
 				    " params structure");
 			alsa_driver_delete (driver);
-			return 0;
+			return NULL;
 		}
 
 		if ((err = snd_pcm_sw_params_malloc (
@@ -2069,7 +2151,7 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 			jack_error ("ALSA: could not allocate playback sw"
 				    " params structure");
 			alsa_driver_delete (driver);
-			return 0;
+			return NULL;
 		}
 	}
 
@@ -2079,7 +2161,7 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 			jack_error ("ALSA: could not allocate capture hw"
 				    " params structure");
 			alsa_driver_delete (driver);
-			return 0;
+			return NULL;
 		}
 
 		if ((err = snd_pcm_sw_params_malloc (
@@ -2087,14 +2169,14 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 			jack_error ("ALSA: could not allocate capture sw"
 				    " params structure");
 			alsa_driver_delete (driver);
-			return 0;
+			return NULL;
 		}
 	}
 
 	if (alsa_driver_set_parameters (driver, frames_per_cycle,
 					user_nperiods, rate)) {
 		alsa_driver_delete (driver);
-		return 0;
+		return NULL;
 	}
 
 	driver->capture_and_playback_not_synced = FALSE;
@@ -2106,8 +2188,6 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 		} 
 	}
 
-	alsa_driver_hw_specific (driver, hw_monitoring, hw_metering);
-
 	driver->client = client;
 
 	return (jack_driver_t *) driver;
@@ -2215,7 +2295,7 @@ driver_get_descriptor ()
 	desc = calloc (1, sizeof (jack_driver_desc_t));
 
 	strcpy (desc->name,"alsa");
-	desc->nparams = 15;
+	desc->nparams = 17;
   
 	params = calloc (desc->nparams, sizeof (jack_driver_param_desc_t));
 
@@ -2266,7 +2346,7 @@ driver_get_descriptor ()
 	params[i].character  = 'n';
 	params[i].type       = JackDriverParamUInt;
 	params[i].value.ui   = 2U;
-	strcpy (params[i].short_desc, "Number of periods in hardware buffer");
+	strcpy (params[i].short_desc, "Number of periods of playback latency");
 	strcpy (params[i].long_desc, params[i].short_desc);
 
 	i++;
@@ -2349,6 +2429,23 @@ driver_get_descriptor ()
 	strcpy (params[i].short_desc, "Try 16-bit samples before 32-bit");
 	strcpy (params[i].long_desc, params[i].short_desc);
 
+
+	i++;
+	strcpy (params[i].name, "input-latency");
+	params[i].character  = 'I';
+	params[i].type       = JackDriverParamUInt;
+	params[i].value.i    = 0;
+	strcpy (params[i].short_desc, "Extra input latency");
+	strcpy (params[i].long_desc, params[i].short_desc);
+
+	i++;
+	strcpy (params[i].name, "output-latency");
+	params[i].character  = 'O';
+	params[i].type       = JackDriverParamUInt;
+	params[i].value.i    = 0;
+	strcpy (params[i].short_desc, "Extra output latency");
+	strcpy (params[i].long_desc, params[i].short_desc);
+
 	desc->params = params;
 
 	return desc;
@@ -2372,6 +2469,8 @@ driver_initialize (jack_client_t *client, const JSList * params)
 	int user_capture_nchnls = 0;
 	int user_playback_nchnls = 0;
 	int shorts_first = FALSE;
+	jack_nframes_t systemic_input_latency = 0;
+	jack_nframes_t systemic_output_latency = 0;
 	const JSList * node;
 	const jack_driver_param_t * param;
 
@@ -2427,6 +2526,8 @@ driver_initialize (jack_client_t *client, const JSList * params)
 				
 		case 'n':
 			user_nperiods = param->value.ui;
+			if (user_nperiods < 2)	/* enforce minimum value */
+				user_nperiods = 2;
 			break;
 				
 		case 's':
@@ -2450,6 +2551,14 @@ driver_initialize (jack_client_t *client, const JSList * params)
 			shorts_first = param->value.i;
 			break;
 
+		case 'I':
+			systemic_input_latency = param->value.ui;
+			break;
+
+		case 'O':
+			systemic_output_latency = param->value.ui;
+			break;
+
 		}
 	}
 			
@@ -2466,7 +2575,9 @@ driver_initialize (jack_client_t *client, const JSList * params)
 				hw_metering, capture, playback, dither,
 				soft_mode, monitor, 
 				user_capture_nchnls, user_playback_nchnls,
-				shorts_first);
+				shorts_first, 
+				systemic_input_latency,
+				systemic_output_latency);
 }
 
 void
