@@ -21,7 +21,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
-    $Id: internal.h,v 1.66 2004/08/27 02:32:11 pbd Exp $
+    $Id: internal.h,v 1.81 2005/05/11 02:06:46 pbd Exp $
 */
 
 #ifndef __jack_internal_h__
@@ -60,6 +60,20 @@ extern void jack_error (const char *fmt, ...);
 #endif
 #endif
 
+/* Enable preemption checking for Linux Realtime Preemption kernels.
+ *
+ * This checks if any RT-safe code section does anything to cause CPU
+ * preemption.  Examples are sleep() or other system calls that block.
+ * If a problem is detected, the kernel writes a syslog entry, and
+ * sends SIGUSR2 to the client.
+ */
+#ifdef DO_PREEMPTION_CHECKING
+#define CHECK_PREEMPTION(engine, onoff) \
+	if ((engine)->real_time) gettimeofday (1, (onoff))
+#else
+#define CHECK_PREEMPTION(engine, onoff)
+#endif
+
 #ifndef	FALSE
 #define	FALSE	(0)
 #endif
@@ -80,10 +94,20 @@ typedef enum {
 } transport_command_t;
 
 typedef struct {
-    volatile jack_time_t guard1;
-    volatile jack_nframes_t frames;
-    volatile jack_time_t stamp;
-    volatile jack_time_t guard2;
+
+	volatile uint32_t       guard1;
+	volatile jack_nframes_t frames;  
+	volatile jack_time_t    current_wakeup;
+	volatile jack_time_t    next_wakeup;
+	volatile float          second_order_integrator;
+	volatile int32_t        initialized;
+	volatile uint32_t       guard2;
+	
+	/* not accessed by clients */
+
+	int32_t  reset_pending;      /* xrun happened, deal with it */
+	float    filter_coefficient; /* set once, never altered */
+
 } jack_frame_timer_t;
 
 /* JACK engine shared memory data structure. */
@@ -106,7 +130,6 @@ typedef struct {
     jack_time_t           sync_time_left;
     jack_frame_timer_t    frame_timer;
     int32_t		  internal;
-    jack_nframes_t        frames_at_cycle_start;
     pid_t                 engine_pid;
     jack_nframes_t	  buffer_size;
     int8_t		  real_time;
@@ -115,6 +138,8 @@ typedef struct {
     int32_t		  client_priority;
     int32_t		  has_capabilities;
     float                 cpu_load;
+    float		  xrun_delayed_usecs;
+    float		  max_delayed_usecs;
     uint32_t		  port_max;
     int32_t		  engine_ok;
     jack_port_type_id_t	  n_port_types;
@@ -135,10 +160,10 @@ typedef enum  {
   XRun,
   StartFreewheel,
   StopFreewheel
-} EventType;
+} JackEventType;
 
 typedef struct {
-    EventType type;
+    JackEventType type;
     union {
 	uint32_t n;
 	jack_port_id_t port_id;
@@ -170,7 +195,7 @@ typedef volatile struct {
     volatile jack_client_id_t id;         /* w: engine r: engine and client */
     volatile jack_nframes_t  nframes;     /* w: engine r: client */
     volatile jack_client_state_t state;   /* w: engine and client r: engine */
-    volatile int8_t     name[JACK_CLIENT_NAME_SIZE];
+    volatile char	name[JACK_CLIENT_NAME_SIZE];
     volatile ClientType type;             /* w: engine r: engine and client */
     volatile int8_t     active;           /* w: engine r: engine and client */
     volatile int8_t     dead;             /* r/w: engine */
@@ -216,7 +241,7 @@ typedef volatile struct {
 	
     /* external clients: set by libjack
      * internal clients: set by engine */
-    int (*deliver_request)(void*, jack_request_t*);
+    int (*deliver_request)(void*, jack_request_t*); /* JOQ: 64/32 bug! */
     void *deliver_arg;
 
     /* for engine use only */
@@ -226,8 +251,10 @@ typedef volatile struct {
 
 typedef struct {
     
+    uint32_t	protocol_v;		/* protocol version, must go first */
     int32_t    load;
     ClientType type;
+    jack_options_t options;
 
     char name[JACK_CLIENT_NAME_SIZE];
     char object_path[PATH_MAX+1];
@@ -237,9 +264,7 @@ typedef struct {
 
 typedef struct {
 
-    int32_t	status;
-
-    uint32_t	protocol_v;
+    jack_status_t status;
 
     jack_shm_info_t client_shm;
     jack_shm_info_t engine_shm;
@@ -249,12 +274,11 @@ typedef struct {
     int32_t	realtime;
     int32_t	realtime_priority;
 
-    /* these two are valid only if the connect request
-       was for type == ClientDriver. 
-    */
+    char name[JACK_CLIENT_NAME_SIZE];	/* unique name, if assigned */
 
-    jack_client_control_t *client_control; /* JOQ: 64/32 problem */
-    jack_control_t        *engine_control; /* JOQ: 64/32 problem */
+    /* these two are valid only for internal clients */
+    jack_client_control_t *client_control;
+    jack_control_t        *engine_control;
 
 #ifdef JACK_USE_MACH_THREADS
     /* specific resources for server/client real-time thread communication */
@@ -290,6 +314,11 @@ typedef enum {
 	SetBufferSize = 16,
 	FreeWheel = 17,
 	StopFreeWheel = 18,
+	IntClientHandle = 19,
+	IntClientLoad = 20,
+	IntClientName = 21,
+	IntClientUnload = 22,
+	RecomputeTotalLatencies = 23
 } RequestType;
 
 struct _jack_request {
@@ -316,15 +345,23 @@ struct _jack_request {
 	    jack_client_id_t client_id;
 	    int32_t conditional;
 	} timebase;
+	struct {
+	    jack_options_t options;
+	    jack_client_id_t id;
+	    char name[JACK_CLIENT_NAME_SIZE];
+	    char path[PATH_MAX+1];
+	    char init[JACK_LOAD_INIT_LIMIT];
+	} intclient;
 	jack_client_id_t client_id;
 	jack_nframes_t nframes;
 	jack_time_t timeout;
+        pid_t cap_pid;
     } x;
     int32_t status;
 };
 
-/* per-client structure allocated in the server's address space
- * its here because its not part of the engine structure.
+/* Per-client structure allocated in the server's address space.
+ * It's here because its not part of the engine structure.
  */
 
 typedef struct _jack_client_internal {
@@ -336,7 +373,10 @@ typedef struct _jack_client_internal {
     int        subgraph_start_fd;
     int        subgraph_wait_fd;
     JSList    *ports;    /* protected by engine->client_lock */
-    JSList    *fed_by;   /* protected by engine->client_lock */
+    JSList    *truefeeds;    /* protected by engine->client_lock */
+    JSList    *sortfeeds;    /* protected by engine->client_lock */
+    int	       fedcount;
+    int	       tfedcount;
     jack_shm_info_t control_shm;
     unsigned long execution_order;
     struct  _jack_client_internal *next_client; /* not a linked list! */
@@ -355,7 +395,14 @@ typedef struct _jack_client_internal {
     
 } jack_client_internal_t;
 
-extern void jack_cleanup_files ();
+typedef struct _jack_thread_arg {
+	jack_client_t* client;
+	void* (*work_function)(void*);
+	int priority;
+	int realtime;
+	void* arg;
+	pid_t cap_pid;
+} jack_thread_arg_t;
 
 extern int  jack_client_handle_port_connection (jack_client_t *client,
 						jack_event_t *event);
@@ -367,7 +414,11 @@ extern jack_client_t *jack_client_alloc_internal (jack_client_control_t*,
 /* internal clients call this. it's defined in jack/engine.c */
 void handle_internal_client_request (jack_control_t*, jack_request_t*);
 
-extern char *jack_server_dir;
+extern char *jack_tmpdir;
+
+extern char *jack_user_dir (void);
+
+extern char *jack_server_dir (const char *server_name);
 
 extern void *jack_zero_filled_buffer;
 
@@ -383,7 +434,16 @@ extern void jack_call_sync_client (jack_client_t *client);
 
 extern void jack_call_timebase_master (jack_client_t *client);
 
+extern char *jack_default_server_name (void);
+
 void silent_jack_error_callback (const char *desc);
 
+/* needed for port management */
+jack_port_t *jack_port_by_id_int (const jack_client_t *client,
+				  jack_port_id_t id, int* free);
+
+jack_port_t *jack_port_by_name_int (jack_client_t *client,
+				    const char *port_name);
+
 #endif /* __jack_internal_h__ */
 
