@@ -1,6 +1,6 @@
 /* -*- mode: c; c-file-style: "bsd"; -*- */
 /*
-    Copyright (C) 2001-2003 Paul Davis
+    Copyright (C) 2001-2005 Paul Davis
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
-    $Id: jackd.c,v 1.41 2004/09/15 02:07:41 joq Exp $
+    $Id: jackd.c,v 1.52 2005/01/16 01:56:45 joq Exp $
 */
 
 #include <config.h>
@@ -38,6 +38,7 @@
 #include <jack/driver.h>
 #include <jack/shm.h>
 #include <jack/driver_parse.h>
+#include <jack/messagebuffer.h>
 
 #ifdef USE_CAPABILITIES
 
@@ -51,9 +52,10 @@ static struct stat pipe_stat;
 
 #endif /* USE_CAPABILITIES */
 
-static JSList * drivers = NULL;
+static JSList *drivers = NULL;
 static sigset_t signals;
-static jack_engine_t *engine = 0;
+static jack_engine_t *engine = NULL;
+static char *server_name = NULL;
 static int realtime = 0;
 static int realtime_priority = 10;
 static int do_mlock = 1;
@@ -137,9 +139,9 @@ jack_main (jack_driver_desc_t * driver_desc, JSList * driver_params)
 	/* get the engine/driver started */
 
 	if ((engine = jack_engine_new (realtime, realtime_priority, 
-				       do_mlock, do_unlock,
-				       temporary, verbose, client_timeout, port_max,
-				       getpid(), drivers)) == 0) {
+				       do_mlock, do_unlock, server_name,
+				       temporary, verbose, client_timeout,
+				       port_max, getpid(), drivers)) == 0) {
 		fprintf (stderr, "cannot create engine\n");
 		return -1;
 	}
@@ -149,12 +151,12 @@ jack_main (jack_driver_desc_t * driver_desc, JSList * driver_params)
 	if (jack_engine_load_driver (engine, driver_desc, driver_params)) {
 		fprintf (stderr, "cannot load driver module %s\n",
 			 driver_desc->name);
-		return -1;
+		goto error;
 	}
 
-        if (engine->driver->start (engine->driver) != 0) {
-                jack_error ("cannot start driver");
-		return -1;
+	if (engine->driver->start (engine->driver) != 0) {
+		jack_error ("cannot start driver");
+		goto error;
 	}
 
 	/* install a do-nothing handler because otherwise pthreads
@@ -196,7 +198,7 @@ jack_main (jack_driver_desc_t * driver_desc, JSList * driver_params)
 			break;
 		}
 	} 
-
+	
 	if (sig != SIGSEGV) {
 
 		/* unblock signals so we can see them during shutdown.
@@ -205,9 +207,13 @@ jack_main (jack_driver_desc_t * driver_desc, JSList * driver_params)
 		*/
 		sigprocmask (SIG_UNBLOCK, &signals, 0);
 	}
-
+	
 	jack_engine_delete (engine);
 	return 1;
+	
+error:
+	jack_engine_delete (engine);
+	return -1;
 }
 
 static jack_driver_desc_t *
@@ -220,10 +226,13 @@ jack_drivers_get_descriptor (JSList * drivers, const char * sofile)
 	char * filename;
 	const char * dlerr;
 	int err;
+	char* driver_dir;
 
-	filename = malloc (strlen (ADDON_DIR) + 1 + strlen (sofile) + 1);
-	sprintf (filename, "%s/%s", ADDON_DIR, sofile);
-
+	if ((driver_dir = getenv("JACK_DRIVER_DIR")) == 0) {
+		driver_dir = ADDON_DIR;
+	}
+	filename = malloc (strlen (driver_dir) + 1 + strlen (sofile) + 1);
+	sprintf (filename, "%s/%s", driver_dir, sofile);
 
 	if (verbose) {
 		fprintf (stderr, "getting driver descriptor from %s\n", filename);
@@ -273,7 +282,6 @@ jack_drivers_get_descriptor (JSList * drivers, const char * sofile)
 	free (filename);
 
 	return descriptor;
-  
 }
 
 static JSList *
@@ -285,12 +293,18 @@ jack_drivers_load ()
 	int err;
 	JSList * driver_list = NULL;
 	jack_driver_desc_t * desc;
+	char* driver_dir;
 
-	/* search through the ADDON_DIR and add get descriptors
+	if ((driver_dir = getenv("JACK_DRIVER_DIR")) == 0) {
+		driver_dir = ADDON_DIR;
+	}
+
+	/* search through the driver_dir and add get descriptors
 	   from the .so files in it */
-	dir_stream = opendir (ADDON_DIR);
+	dir_stream = opendir (driver_dir);
 	if (!dir_stream) {
-		jack_error ("could not open driver directory %s: %s\n", ADDON_DIR, strerror (errno));
+		jack_error ("could not open driver directory %s: %s\n",
+			    driver_dir, strerror (errno));
 		return NULL;
 	}
   
@@ -313,16 +327,16 @@ jack_drivers_load ()
 		if (desc) {
 			driver_list = jack_slist_append (driver_list, desc);
 		}
-
 	}
 
 	err = closedir (dir_stream);
 	if (err) {
-		jack_error ("error closing driver directory %s: %s\n", ADDON_DIR, strerror (errno));
+		jack_error ("error closing driver directory %s: %s\n",
+			    driver_dir, strerror (errno));
 	}
 
 	if (!driver_list) {
-		jack_error ("could not find any drivers in %s!\n", ADDON_DIR);
+		jack_error ("could not find any drivers in %s!\n", driver_dir);
 		return NULL;
 	}
 
@@ -332,7 +346,7 @@ jack_drivers_load ()
 static void copyright (FILE* file)
 {
 	fprintf (file, "jackd " VERSION "\n"
-"Copyright 2001-2003 Paul Davis and others.\n"
+"Copyright 2001-2005 Paul Davis and others.\n"
 "jackd comes with ABSOLUTELY NO WARRANTY\n"
 "This is free software, and you are welcome to redistribute it\n"
 "under certain conditions; see the file COPYING for details\n\n");
@@ -343,6 +357,7 @@ static void usage (FILE *file)
 	copyright (file);
 	fprintf (file, "\n"
 "usage: jackd [ --realtime OR -R [ --realtime-priority OR -P priority ] ]\n"
+"             [ --name OR -n server-name ]\n"
 "             [ --no-mlock OR -m ]\n"
 "             [ --unlock OR -u ]\n"
 "             [ --timeout OR -t client-timeout-in-msecs ]\n"
@@ -376,12 +391,76 @@ jack_find_driver_descriptor (const char * name)
 	return desc;
 }
 
+static void
+jack_cleanup_files (const char *server_name)
+{
+	DIR *dir;
+	struct dirent *dirent;
+	char *dir_name = jack_server_dir (server_name);
+
+	/* On termination, we remove all files that jackd creates so
+	 * subsequent attempts to start jackd will not believe that an
+	 * instance is already running.  If the server crashes or is
+	 * terminated with SIGKILL, this is not possible.  So, cleanup
+	 * is also attempted when jackd starts.
+	 *
+	 * There are several tricky issues.  First, the previous JACK
+	 * server may have run for a different user ID, so its files
+	 * may be inaccessible.  This is handled by using a separate
+	 * JACK_TMP_DIR subdirectory for each user.  Second, there may
+	 * be other servers running with different names.  Each gets
+	 * its own subdirectory within the per-user directory.  The
+	 * current process has already registered as `server_name', so
+	 * we know there is no other server actively using that name.
+	 */
+
+	/* nothing to do if the server directory does not exist */
+	if ((dir = opendir (dir_name)) == NULL) {
+		return;
+	}
+
+	/* unlink all the files in this directory, they are mine */
+	while ((dirent = readdir (dir)) != NULL) {
+
+		char fullpath[PATH_MAX];
+
+		if ((strcmp (dirent->d_name, ".") == 0)
+		    || (strcmp (dirent->d_name, "..") == 0)) {
+			continue;
+		}
+
+		snprintf (fullpath, sizeof (fullpath), "%s/%s",
+			  dir_name, dirent->d_name);
+
+		if (unlink (fullpath)) {
+			jack_error ("cannot unlink `%s' (%s)", fullpath,
+				    strerror (errno));
+		}
+	} 
+
+	closedir (dir);
+
+	/* now, delete the per-server subdirectory, itself */
+	if (rmdir (dir_name)) {
+ 		jack_error ("cannot remove `%s' (%s)", dir_name,
+			    strerror (errno));
+	}
+
+	/* finally, delete the per-user subdirectory, if empty */
+	if (rmdir (jack_user_dir ())) {
+		if (errno != ENOTEMPTY) {
+			jack_error ("cannot remove `%s' (%s)",
+				    jack_user_dir (), strerror (errno));
+		}
+	}
+}
+
 int	       
 main (int argc, char *argv[])
 
 {
 	jack_driver_desc_t * desc;
-	const char *options = "-ad:P:vshVRTFl:t:mp:";
+	const char *options = "-ad:P:uvshVRTFl:t:mn:p:";
 	struct option long_options[] = 
 	{ 
 		{ "driver", 1, 0, 'd' },
@@ -389,7 +468,8 @@ main (int argc, char *argv[])
 		{ "help", 0, 0, 'h' },
 		{ "port-max", 1, 0, 'p' },
 		{ "no-mlock", 0, 0, 'm' },
-		{ "unmlock", 0, 0, 'u' },
+		{ "name", 0, 0, 'n' },
+		{ "unlock", 0, 0, 'u' },
 		{ "realtime", 0, 0, 'R' },
 		{ "realtime-priority", 1, 0, 'P' },
 		{ "timeout", 1, 0, 't' },
@@ -401,12 +481,13 @@ main (int argc, char *argv[])
 	int opt = 0;
 	int option_index = 0;
 	int seen_driver = 0;
-	char *driver_name = 0;
-	char **driver_args = 0;
+	char *driver_name = NULL;
+	char **driver_args = NULL;
 	JSList * driver_params;
 	int driver_nargs = 1;
 	int show_version = 0;
 	int i;
+	int rc;
 #ifdef USE_CAPABILITIES
 	int status;
 #endif
@@ -466,6 +547,10 @@ main (int argc, char *argv[])
 			jack_set_error_function (silent_jack_error_callback);
 			break;
 
+		case 'n':
+			server_name = optarg;
+			break;
+
 		case 'm':
 			do_mlock = 0;
 			break;
@@ -499,7 +584,8 @@ main (int argc, char *argv[])
 			break;
 
 		default:
-			fprintf (stderr, "unknown option character %c\n", optopt);
+			fprintf (stderr, "unknown option character %c\n",
+				 optopt);
 			/*fallthru*/
 		case 'h':
 			usage (stdout);
@@ -508,17 +594,10 @@ main (int argc, char *argv[])
 	}
 
 	if (show_version) {
-#ifdef DEFAULT_TMP_DIR
 		printf ( "jackd version " VERSION 
 				" tmpdir " DEFAULT_TMP_DIR 
 				" protocol " PROTOCOL_VERSION
 				"\n");
-#else
-		printf ( "jackd version " VERSION 
-				" tmpdir " "/tmp" 
-				" protocol " PROTOCOL_VERSION
-				"\n");
-#endif
 		return -1;
 	}
 
@@ -528,15 +607,13 @@ main (int argc, char *argv[])
 	}
 
 	drivers = jack_drivers_load ();
-	if (!drivers)
-	{
+	if (!drivers) {
 		fprintf (stderr, "jackd: no drivers found; exiting\n");
 		exit (1);
 	}
 
 	desc = jack_find_driver_descriptor (driver_name);
-	if (!desc)
-	{
+	if (!desc) {
 		fprintf (stderr, "jackd: unknown driver '%s'\n", driver_name);
 		exit (1);
 	}
@@ -560,20 +637,51 @@ main (int argc, char *argv[])
 		driver_args[i] = argv[optind++];
 	}
 
-	i = jack_parse_driver_params (desc, driver_nargs, driver_args, &driver_params);
-	if (i)
+	if (jack_parse_driver_params (desc, driver_nargs,
+				      driver_args, &driver_params)) {
 		exit (0);
+	}
+
+	if (server_name == NULL)
+		server_name = jack_default_server_name ();
 
 	copyright (stdout);
 
-	jack_cleanup_shm ();
-	jack_cleanup_files ();
+	rc = jack_register_server (server_name);
+	switch (rc) {
+	case EEXIST:
+		fprintf (stderr, "`%s' server already active\n", server_name);
+		exit (1);
+	case ENOSPC:
+		fprintf (stderr, "too many servers already active\n");
+		exit (2);
+	case ENOMEM:
+		fprintf (stderr, "no access to shm registry\n");
+		exit (3);
+	default:
+		if (verbose)
+			fprintf (stderr, "server `%s' registered\n",
+				 server_name);
+	}
 
+	/* clean up shared memory and files from any previous
+	 * instance of this server name */
+	jack_cleanup_shm ();
+	jack_cleanup_files (server_name);
+
+	/* run the server engine until it terminates */
 	jack_main (desc, driver_params);
 
+	/* clean up shared memory and files from this server instance */
+	if (verbose)
+		fprintf (stderr, "cleaning up shared memory\n");
 	jack_cleanup_shm ();
-	jack_cleanup_files ();
+	if (verbose)
+		fprintf (stderr, "cleaning up files\n");
+	jack_cleanup_files (server_name);
+	if (verbose)
+		fprintf (stderr, "unregistering server `%s'\n", server_name);
+	jack_unregister_server (server_name);
 
 	exit (0);
 }
-
