@@ -16,7 +16,7 @@
     along with this program; if not, write to the Free Software 
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 
-    $Id: client.c,v 1.91 2004/09/15 01:33:06 joq Exp $
+    $Id: client.c,v 1.121 2005/06/10 14:00:20 trutkin Exp $
 */
 
 #include <config.h>
@@ -45,6 +45,7 @@
 #include <jack/shm.h>
 #include <jack/unlock.h>
 #include <jack/thread.h>
+#include <jack/varargs.h>
 
 #include <sysdeps/time.h>
 JACK_TIME_GLOBAL_DECL;			/* One instance per process. */
@@ -62,16 +63,6 @@ JACK_TIME_GLOBAL_DECL;			/* One instance per process. */
 #include <jack/timestamps.h>
 #endif /* WITH_TIMESTAMPS */
 
-char *jack_server_dir = DEFAULT_TMP_DIR;
-
-void
-jack_set_server_dir (const char *path)
-{
-	jack_error ("jack_set_server_dir() is deprecated.\n"
-		    "Please contact the program's author");
-	jack_server_dir = strdup (path);
-}
-
 static pthread_mutex_t client_lock;
 static pthread_cond_t  client_ready;
 void *jack_zero_filled_buffer = NULL;
@@ -123,8 +114,9 @@ oop_client_deliver_request (void *ptr, jack_request_t *req)
 	rok = (read (client->request_fd, req, sizeof (*req))
 	       == sizeof (*req));
 
-	if (wok && rok)			/* everything OK? */
+	if (wok && rok) {		/* everything OK? */
 		return req->status;
+	}
 
 	req->status = -1;		/* request failed */
 
@@ -145,14 +137,45 @@ oop_client_deliver_request (void *ptr, jack_request_t *req)
 int
 jack_client_deliver_request (const jack_client_t *client, jack_request_t *req)
 {
-	/* indirect through the function pointer that was set 
-	   either by jack_client_new() (external) or handle_new_client()
-	   in the server.
-	*/
+	/* indirect through the function pointer that was set either
+	 * by jack_client_open() or by jack_new_client_request() in
+	 * the server.
+	 */
 
-	return client->control->deliver_request (client->control->deliver_arg, req);
+	return client->control->deliver_request (client->control->deliver_arg,
+						 req);
 }
 
+#if JACK_USE_MACH_THREADS 
+
+jack_client_t *
+jack_client_alloc ()
+{
+	jack_client_t *client;
+
+	client = (jack_client_t *) malloc (sizeof (jack_client_t));
+	client->pollfd = (struct pollfd *) malloc (sizeof (struct pollfd) * 1);
+	client->pollmax = 1;
+	client->request_fd = -1;
+	client->event_fd = -1;
+	client->upstream_is_jackd = 0;
+	client->graph_wait_fd = -1;
+	client->graph_next_fd = -1;
+	client->ports = NULL;
+	client->ports_ext = NULL;
+	client->engine = NULL;
+	client->control = NULL;
+	client->thread_ok = FALSE;
+	client->rt_thread_ok = FALSE;
+	client->first_active = TRUE;
+	client->on_shutdown = NULL;
+	client->n_port_types = 0;
+	client->port_segment = NULL;
+	return client;
+}
+
+#else
+
 jack_client_t *
 jack_client_alloc ()
 {
@@ -167,12 +190,10 @@ jack_client_alloc ()
 	client->graph_wait_fd = -1;
 	client->graph_next_fd = -1;
 	client->ports = NULL;
+	client->ports_ext = NULL;
 	client->engine = NULL;
 	client->control = NULL;
 	client->thread_ok = FALSE;
-#if JACK_USE_MACH_THREADS 
-	client->rt_thread_ok = FALSE;
-#endif
 	client->first_active = TRUE;
 	client->on_shutdown = NULL;
 	client->n_port_types = 0;
@@ -181,6 +202,8 @@ jack_client_alloc ()
 	return client;
 }
 
+#endif
+
 /*
  * Build the jack_client_t structure for an internal client.
  */
@@ -236,32 +259,44 @@ jack_client_invalidate_port_buffers (jack_client_t *client)
 
 int
 jack_client_handle_port_connection (jack_client_t *client, jack_event_t *event)
-
 {
 	jack_port_t *control_port;
 	jack_port_t *other;
 	JSList *node;
+	int need_free = FALSE;
 
 	switch (event->type) {
 	case PortConnected:
-		other = jack_port_new (client, event->y.other_id, client->engine);
-		control_port = jack_port_by_id (client, event->x.self_id);
+		other = jack_port_new (client, event->y.other_id,
+				       client->engine);
+		/* jack_port_by_id_int() always returns an internal
+		 * port that does not need to be deallocated */
+		control_port = jack_port_by_id_int (client, event->x.self_id,
+						    &need_free);
 		pthread_mutex_lock (&control_port->connection_lock);
-		control_port->connections = jack_slist_prepend (control_port->connections, (void*)other);
+		control_port->connections =
+			jack_slist_prepend (control_port->connections,
+					    (void *) other);
 		pthread_mutex_unlock (&control_port->connection_lock);
 		break;
 
 	case PortDisconnected:
-		control_port = jack_port_by_id (client, event->x.self_id);
-
+		/* jack_port_by_id_int() always returns an internal
+		 * port that does not need to be deallocated */
+		control_port = jack_port_by_id_int (client, event->x.self_id,
+						    &need_free);
 		pthread_mutex_lock (&control_port->connection_lock);
 
-		for (node = control_port->connections; node; node = jack_slist_next (node)) {
+		for (node = control_port->connections; node;
+		     node = jack_slist_next (node)) {
 
 			other = (jack_port_t *) node->data;
 
 			if (other->shared->id == event->y.other_id) {
-				control_port->connections = jack_slist_remove_link (control_port->connections, node);
+				control_port->connections =
+					jack_slist_remove_link (
+						control_port->connections,
+						node);
 				jack_slist_free_1 (node);
 				free (other);
 				break;
@@ -279,6 +314,26 @@ jack_client_handle_port_connection (jack_client_t *client, jack_event_t *event)
 	return 0;
 }
 
+#if JACK_USE_MACH_THREADS
+
+static int 
+jack_handle_reorder (jack_client_t *client, jack_event_t *event)
+{	
+	client->pollmax = 1;
+
+	/* If the client registered its own callback for graph order events,
+	   execute it now.
+	*/
+
+	if (client->control->graph_order) {
+		client->control->graph_order (client->control->graph_order_arg);
+	}
+
+	return 0;
+}
+
+#else
+
 static int 
 jack_handle_reorder (jack_client_t *client, jack_event_t *event)
 {	
@@ -297,25 +352,27 @@ jack_handle_reorder (jack_client_t *client, jack_event_t *event)
 	}
 
 	sprintf (path, "%s-%" PRIu32, client->fifo_prefix, event->x.n);
-
+	
 	if ((client->graph_wait_fd = open (path, O_RDONLY|O_NONBLOCK)) < 0) {
-		jack_error ("cannot open specified fifo [%s] for reading (%s)", path, strerror (errno));
+		jack_error ("cannot open specified fifo [%s] for reading (%s)",
+			    path, strerror (errno));
 		return -1;
 	}
-
 	DEBUG ("opened new graph_wait_fd %d (%s)", client->graph_wait_fd, path);
 
 	sprintf (path, "%s-%" PRIu32, client->fifo_prefix, event->x.n+1);
 	
 	if ((client->graph_next_fd = open (path, O_WRONLY|O_NONBLOCK)) < 0) {
-		jack_error ("cannot open specified fifo [%s] for writing (%s)", path, strerror (errno));
+		jack_error ("cannot open specified fifo [%s] for writing (%s)",
+			    path, strerror (errno));
 		return -1;
 	}
 
 	client->upstream_is_jackd = event->y.n;
 	client->pollmax = 2;
 
-	DEBUG ("opened new graph_next_fd %d (%s) (upstream is jackd? %d)", client->graph_next_fd, path, 
+	DEBUG ("opened new graph_next_fd %d (%s) (upstream is jackd? %d)",
+	       client->graph_next_fd, path, 
 	       client->upstream_is_jackd);
 
 	/* If the client registered its own callback for graph order events,
@@ -328,12 +385,15 @@ jack_handle_reorder (jack_client_t *client, jack_event_t *event)
 
 	return 0;
 }
+
+#endif
 		
 static int
-server_connect (int which)
+server_connect (const char *server_name)
 {
 	int fd;
 	struct sockaddr_un addr;
+	int which = 0;
 
 	if ((fd = socket (AF_UNIX, SOCK_STREAM, 0)) < 0) {
 		jack_error ("cannot create client socket (%s)",
@@ -341,9 +401,12 @@ server_connect (int which)
 		return -1;
 	}
 
+	//JOQ: temporary debug message
+	//fprintf (stderr, "DEBUG: connecting to `%s' server\n", server_name);
+
 	addr.sun_family = AF_UNIX;
-	snprintf (addr.sun_path, sizeof (addr.sun_path) - 1, "%s/jack_%d_%d",
-		  jack_server_dir, getuid (), which);
+	snprintf (addr.sun_path, sizeof (addr.sun_path) - 1, "%s/jack_%d",
+		  jack_server_dir (server_name), which);
 
 	if (connect (fd, (struct sockaddr *) &addr, sizeof (addr)) < 0) {
 		close (fd);
@@ -354,7 +417,7 @@ server_connect (int which)
 }
 
 static int
-server_event_connect (jack_client_t *client)
+server_event_connect (jack_client_t *client, const char *server_name)
 {
 	int fd;
 	struct sockaddr_un addr;
@@ -368,8 +431,8 @@ server_event_connect (jack_client_t *client)
 	}
 
 	addr.sun_family = AF_UNIX;
-	snprintf (addr.sun_path, sizeof (addr.sun_path) - 1, "%s/jack_%d_ack_0",
-		  jack_server_dir, getuid () );
+	snprintf (addr.sun_path, sizeof (addr.sun_path) - 1, "%s/jack_ack_0",
+		  jack_server_dir (server_name));
 
 	if (connect (fd, (struct sockaddr *) &addr, sizeof (addr)) < 0) {
 		jack_error ("cannot connect to jack server for events",
@@ -406,7 +469,7 @@ server_event_connect (jack_client_t *client)
 
 /* Exec the JACK server in this process.  Does not return. */
 static void
-_start_server (void)
+_start_server (const char *server_name)
 {
 	FILE* fp = 0;
 	char filename[255];
@@ -419,8 +482,7 @@ _start_server (void)
 	int i = 0;
 	int good = 0;
 	int ret;
-
-
+	
 	snprintf(filename, 255, "%s/.jackdrc", getenv("HOME"));
 	fp = fopen(filename, "r");
 
@@ -453,20 +515,27 @@ _start_server (void)
 #endif /* USE_CAPABILITIES */
 	} else {
 		result = strcspn(arguments, " ");
-		command = (char*)malloc(result+1);
+		command = (char *) malloc(result+1);
 		strncpy(command, arguments, result);
 		command[result] = '\0';
 	}
 
-	argv = (char**)malloc(255);
+	argv = (char **) malloc (255);
   
 	while(1) {
-		/* insert -T into arguments */
+		/* insert -T and -nserver_name in front of arguments */
 		if (i == 1) {
-			argv[i] = (char*)malloc(3);
-			strncpy(argv[i], "-T", 2); 
-			argv[i][2] = '\0';
-			++i;
+			argv[i] = (char *) malloc(strlen ("-T") + 1);
+			strcpy (argv[i++], "-T"); 
+			if (server_name) {
+				size_t optlen = strlen ("-n");
+				char *buf =
+					malloc (optlen
+						+ strlen (server_name) + 1);
+				strcpy (buf, "-n");
+				strcpy (buf+optlen, server_name);
+				argv[i++] = buf;
+			}
 		}
 
 		result = strcspn(arguments+pos, " ");
@@ -482,7 +551,6 @@ _start_server (void)
 		++i;
 	}
 	argv[i] = 0;
-
 	execv (command, argv);
 
 	/* If execv() succeeds, it does not return.  There's no point
@@ -491,19 +559,17 @@ _start_server (void)
 }
 
 int
-start_server (void)
+start_server (const char *server_name, jack_options_t options)
 {
-	/* Only fork() a server when $JACK_START_SERVER is defined and
-	 * $JACK_NO_START_SERVER is not. */
-	if (getenv("JACK_START_SERVER") == NULL ||
-	    getenv("JACK_NO_START_SERVER") != NULL) {
+	if ((options & JackNoStartServer)
+	    || getenv("JACK_NO_START_SERVER")) {
 		return 1;
 	}
 
 	/* The double fork() forces the server to become a child of
 	 * init, which will always clean up zombie process state on
-	 * termination.  This even works in strange corner cases where
-	 * the server terminates but this client does not.
+	 * termination.  This even works in cases where the server
+	 * terminates but this client does not.
 	 *
 	 * Since fork() is usually implemented using copy-on-write
 	 * virtual memory tricks, the overhead of the second fork() is
@@ -513,7 +579,7 @@ start_server (void)
 	case 0:				/* child process */
 		switch (fork()) {
 		case 0:			/* grandchild process */
-			_start_server();
+			_start_server(server_name);
 			_exit (99);	/* exec failed */
 		case -1:
 			_exit (98);
@@ -529,15 +595,16 @@ start_server (void)
 }
 
 static int
-jack_request_client (ClientType type, const char* client_name,
-		     const char* so_name, const char* so_data,
+jack_request_client (ClientType type,
+		     const char* client_name, jack_options_t options,
+		     jack_status_t *status, jack_varargs_t *va,
 		     jack_client_connect_result_t *res, int *req_fd)
 {
 	jack_client_connect_request_t req;
 
 	*req_fd = -1;
-
 	memset (&req, 0, sizeof (req));
+	req.options = options;
 
 	if (strlen (client_name) >= sizeof (req.name)) {
 		jack_error ("\"%s\" is too long to be used as a JACK client"
@@ -547,45 +614,56 @@ jack_request_client (ClientType type, const char* client_name,
 		return -1;
 	}
 
-	if (strlen (so_name) > sizeof (req.object_path) - 1) {
+	if (va->load_name
+	    && (strlen (va->load_name) > sizeof (req.object_path) - 1)) {
 		jack_error ("\"%s\" is too long to be used as a JACK shared"
 			    " object name.\n"
 			     "Please use %lu characters or less.",
-			    so_name, sizeof (req.object_path) - 1);
+			    va->load_name, sizeof (req.object_path) - 1);
 		return -1;
 	}
 
-	if (strlen (so_data) > sizeof (req.object_data) - 1) {
+	if (va->load_init
+	    && (strlen (va->load_init) > sizeof (req.object_data) - 1)) {
 		jack_error ("\"%s\" is too long to be used as a JACK shared"
 			    " object data string.\n"
 			     "Please use %lu characters or less.",
-			    so_data, sizeof (req.object_data) - 1);
+			    va->load_init, sizeof (req.object_data) - 1);
 		return -1;
 	}
 
-	if ((*req_fd = server_connect (0)) < 0) {
+	if ((*req_fd = server_connect (va->server_name)) < 0) {
 		int trys;
-		if (start_server()) {
+		if (start_server(va->server_name, options)) {
+			*status |= (JackFailure|JackServerFailed);
 			goto fail;
 		}
 		trys = 5;
 		do {
 			sleep(1);
 			if (--trys < 0) {
+				*status |= (JackFailure|JackServerFailed);
 				goto fail;
 			}
-		} while ((*req_fd = server_connect (0)) < 0);
+		} while ((*req_fd = server_connect (va->server_name)) < 0);
+		*status |= JackServerStarted;
 	}
 
+	/* format connection request */
+	req.protocol_v = jack_protocol_version;
 	req.load = TRUE;
 	req.type = type;
-	snprintf (req.name, sizeof (req.name), "%s", client_name);
-	snprintf (req.object_path, sizeof (req.object_path), "%s", so_name);
-	snprintf (req.object_data, sizeof (req.object_data), "%s", so_data);
+	snprintf (req.name, sizeof (req.name),
+		  "%s", client_name);
+	snprintf (req.object_path, sizeof (req.object_path),
+		  "%s", va->load_name);
+	snprintf (req.object_data, sizeof (req.object_data),
+		  "%s", va->load_init);
 
 	if (write (*req_fd, &req, sizeof (req)) != sizeof (req)) {
 		jack_error ("cannot send request to jack server (%s)",
 			    strerror (errno));
+		*status |= (JackFailure|JackServerError);
 		goto fail;
 	}
 
@@ -593,25 +671,33 @@ jack_request_client (ClientType type, const char* client_name,
 
 		if (errno == 0) {
 			/* server shut the socket */
-			jack_error ("could not attach as client "
-				    "(duplicate client name?)");
+			jack_error ("could not attach as client");
+			*status |= (JackFailure|JackServerError);
 			goto fail;
 		}
-
+		
+		if (errno == ECONNRESET) {
+			jack_error ("could not attach as JACK client "
+				    "(server has exited)");
+			*status |= (JackFailure|JackServerError);
+			goto fail;
+		}
+		
 		jack_error ("cannot read response from jack server (%s)",
 			    strerror (errno));
+		*status |= (JackFailure|JackServerError);
 		goto fail;
 	}
 
-	if (res->status) {
-		jack_error ("could not attach as client "
-			    "(duplicate client name?)");
-		goto fail;
-	}
+	*status |= res->status;		/* return server status bits */
 
-	if (res->protocol_v != jack_protocol_version){
-		jack_error ("application linked against incompatible libjack"
-			    " version.");
+	if (*status & JackFailure) {
+		if (*status & JackVersionError) {
+			jack_error ("client linked with incompatible libjack"
+				    " version.");
+		}
+		jack_error ("could not attach to JACK server");
+		*status |= JackServerError;
 		goto fail;
 	}
 
@@ -668,15 +754,7 @@ jack_attach_port_segment (jack_client_t *client, jack_port_type_id_t ptid)
 	} else {
 
 		/* release any previous segment */
-		
-#if JACK_USE_MACH_THREADS 
-		/* Stephane Letz : letz@grame.fr
-		Need a fix : this crash on MacOSX : temporary removed 
 		jack_release_shm (&client->port_segment[ptid]);
-		*/
-#else
-		jack_release_shm (&client->port_segment[ptid]);
-#endif
 	}
 
 	/* get the index into the shm registry */
@@ -707,36 +785,65 @@ jack_attach_port_segment (jack_client_t *client, jack_port_type_id_t ptid)
 }
 
 jack_client_t *
-jack_client_new (const char *client_name)
+jack_client_open (const char *client_name,
+		  jack_options_t options,
+		  jack_status_t *status, ...)
 {
+	/* optional arguments: */
+	va_list ap;			/* variable argument pointer */
+	jack_varargs_t va;		/* variable arguments */
+
 	int req_fd = -1;
 	int ev_fd = -1;
 	jack_client_connect_result_t  res;
 	jack_client_t *client;
 	jack_port_type_id_t ptid;
+	jack_status_t my_status;
 
-	/* external clients need this initialized; internal clients
-	   will use the setup in the server's address space.
-	*/
+	if (status == NULL)		/* no status from caller? */
+		status = &my_status;	/* use local status word */
+	*status = 0;
+
+	/* validate parameters */
+	if ((options & ~JackOpenOptions)) {
+		*status |= (JackFailure|JackInvalidOption);
+		return NULL;
+	}
+
+	/* parse variable arguments */
+	va_start (ap, status);
+	jack_varargs_parse (options, ap, &va);
+	va_end (ap);
+
+	/* External clients need this initialized.  It is already set
+	 * up in the server's address space for internal clients.
+	 */
 	jack_init_time ();
 
-	if (jack_initialize_shm ()) {
-		jack_error ("Unable to initialize shared memory.");
-		return NULL;
-	}
-
-	if (jack_request_client (ClientExternal, client_name, "", "",
-				 &res, &req_fd)) {
+	if (jack_request_client (ClientExternal, client_name, options, status,
+				 &va, &res, &req_fd)) {
 		return NULL;
 	}
 
+	/* Allocate the jack_client_t structure in local memory.
+	 * Shared memory is not accessible yet. */
 	client = jack_client_alloc ();
-
+	strcpy (client->name, res.name);
 	strcpy (client->fifo_prefix, res.fifo_prefix);
 	client->request_fd = req_fd;
+	client->pollfd[EVENT_POLL_INDEX].events =
+		POLLIN|POLLERR|POLLHUP|POLLNVAL;
+#ifndef JACK_USE_MACH_THREADS
+	client->pollfd[WAIT_POLL_INDEX].events =
+		POLLIN|POLLERR|POLLHUP|POLLNVAL;
+#endif
 
-	client->pollfd[EVENT_POLL_INDEX].events = POLLIN|POLLERR|POLLHUP|POLLNVAL;
-	client->pollfd[WAIT_POLL_INDEX].events = POLLIN|POLLERR|POLLHUP|POLLNVAL;
+	/* Don't access shared memory until server connected. */
+	if (jack_initialize_shm (va.server_name)) {
+		jack_error ("Unable to initialize shared memory.");
+		*status |= (JackFailure|JackShmFailure);
+		goto fail;
+	}
 
 	/* attach the engine control/info block */
 	client->engine_shm = res.engine_shm;
@@ -759,10 +866,10 @@ jack_client_new (const char *client_name)
 	client->control = (jack_client_control_t *)
 		jack_shm_addr (&client->control_shm);
 
-	/* nobody else needs to access this shared memory any more, so
-	   destroy it. because we have our own attachment to it, it won't
-	   vanish till we exit (and release it).
-	*/
+	/* Nobody else needs to access this shared memory any more, so
+	 * destroy it.  Because we have it attached, it won't vanish
+	 * till we exit (and release it).
+	 */
 	jack_destroy_shm (&client->control_shm);
 
 	client->n_port_types = client->engine->n_port_types;
@@ -782,7 +889,7 @@ jack_client_new (const char *client_name)
 	client->control->deliver_request = oop_client_deliver_request;
 	client->control->deliver_arg = client;
 
-	if ((ev_fd = server_event_connect (client)) < 0) {
+	if ((ev_fd = server_event_connect (client, va.server_name)) < 0) {
 		goto fail;
 	}
 
@@ -820,17 +927,87 @@ jack_client_new (const char *client_name)
 	if (ev_fd >= 0) {
 		close (ev_fd);
 	}
+	free (client);
 
-	return 0;
+	return NULL;
+}
+
+jack_client_t *
+jack_client_new (const char *client_name)
+{
+	jack_options_t options = JackUseExactName;
+	if (getenv("JACK_START_SERVER") == NULL)
+		options |= JackNoStartServer;
+
+	return jack_client_open (client_name, options, NULL);
+}
+
+char *
+jack_get_client_name (jack_client_t *client)
+{
+	return client->name;
 }
 
 int
-jack_internal_client_new (const char *client_name, const char *so_name, const char *so_data)
+jack_internal_client_new (const char *client_name,
+			  const char *so_name, const char *so_data)
 {
 	jack_client_connect_result_t res;
 	int req_fd;
-	
-	return jack_request_client (ClientInternal, client_name, so_name, so_data, &res, &req_fd);
+	jack_varargs_t va;
+	jack_status_t status;
+	jack_options_t options = JackUseExactName;
+
+	if (getenv("JACK_START_SERVER") == NULL)
+		options |= JackNoStartServer;
+
+	jack_varargs_init (&va);
+	va.load_name = (char *) so_name;
+	va.load_init = (char *) so_data;
+
+	return jack_request_client (ClientInternal, client_name,
+				    options, &status, &va, &res, &req_fd);
+}
+
+char *
+jack_default_server_name (void)
+{
+	char *server_name;
+	if ((server_name = getenv("JACK_DEFAULT_SERVER")) == NULL)
+		server_name = "default";
+	return server_name;
+}
+
+char *jack_tmpdir = DEFAULT_TMP_DIR;
+
+/* returns the name of the per-user subdirectory of jack_tmpdir */
+char *
+jack_user_dir (void)
+{
+	static char user_dir[PATH_MAX] = "";
+
+	/* format the path name on the first call */
+	if (user_dir[0] == '\0') {
+		snprintf (user_dir, sizeof (user_dir), "%s/jack-%d",
+			  jack_tmpdir, getuid ());
+	}
+
+	return user_dir;
+}
+
+/* returns the name of the per-server subdirectory of jack_user_dir() */
+char *
+jack_server_dir (const char *server_name)
+{
+	static char server_dir[PATH_MAX] = "";
+
+	/* format the path name on the first call */
+	if (server_dir[0] == '\0') {
+		snprintf (server_dir, sizeof (server_dir), "%s/%s",
+			  jack_user_dir (), server_name);
+	}
+
+	return server_dir;
 }
 
 void
@@ -838,16 +1015,18 @@ jack_internal_client_close (const char *client_name)
 {
 	jack_client_connect_request_t req;
 	int fd;
+	char *server_name = jack_default_server_name ();
 
 	req.load = FALSE;
 	snprintf (req.name, sizeof (req.name), "%s", client_name);
 	
-	if ((fd = server_connect (0)) < 0) {
+	if ((fd = server_connect (server_name)) < 0) {
 		return;
 	}
 
 	if (write (fd, &req, sizeof (req)) != sizeof(req)) {
-		jack_error ("cannot deliver ClientUnload request to JACK server.");
+		jack_error ("cannot deliver ClientUnload request to JACK "
+			    "server.");
 	}
 	
 	/* no response to this request */
@@ -856,11 +1035,20 @@ jack_internal_client_close (const char *client_name)
 	return;
 }
 
+int
+jack_recompute_total_latencies (jack_client_t* client)
+{
+	jack_request_t request;
+
+	request.type = RecomputeTotalLatencies;
+	return jack_client_deliver_request (client, &request);
+}
 
 int
 jack_set_freewheel (jack_client_t* client, int onoff)
 {
 	jack_request_t request;
+
 	request.type = onoff ? FreeWheel : StopFreeWheel;
 	return jack_client_deliver_request (client, &request);
 }
@@ -888,19 +1076,19 @@ jack_stop_freewheel (jack_client_t* client)
 {
 	jack_client_control_t *control = client->control;
 
-	if (control->freewheel_cb) {
-		control->freewheel_cb (0, control->freewheel_arg);
-	}
-
 	if (client->engine->real_time) {
 #if JACK_USE_MACH_THREADS 
 		jack_acquire_real_time_scheduling (client->process_thread,
-						   client->engine->client_priority);
+				client->engine->client_priority);
 #else
-		jack_acquire_real_time_scheduling (client->thread,
-						   client->engine->client_priority);
+		jack_acquire_real_time_scheduling (client->thread, 
+				client->engine->client_priority);
 #endif
 	}
+
+	if (control->freewheel_cb) {
+		control->freewheel_cb (0, control->freewheel_arg);
+	}
 }
 
 static void *
@@ -910,8 +1098,10 @@ jack_client_thread (void *arg)
 	jack_client_control_t *control = client->control;
 	jack_event_t event;
 	char status = 0;
-	char c = 0;
 	int err = 0;
+#ifndef JACK_USE_MACH_THREADS                   
+	char c = 0;
+#endif
 
 	pthread_mutex_lock (&client_lock);
 	client->thread_ok = TRUE;
@@ -958,20 +1148,25 @@ jack_client_thread (void *arg)
 		 * process() cycle. 
 		 */
 
-		if (client->graph_wait_fd >= 0 && client->pollfd[WAIT_POLL_INDEX].revents & POLLIN) {
+	#ifndef JACK_USE_MACH_THREADS 
+		if (client->graph_wait_fd >= 0
+		    && client->pollfd[WAIT_POLL_INDEX].revents & POLLIN) {
 			control->awake_at = jack_get_microseconds();
 		}
-
+	
 		DEBUG ("pfd[EVENT].revents = 0x%x pfd[WAIT].revents = 0x%x",
 		       client->pollfd[EVENT_POLL_INDEX].revents,
 		       client->pollfd[WAIT_POLL_INDEX].revents);
-
+	#endif
+	
 		pthread_testcancel();
 		
+	#ifndef JACK_USE_MACH_THREADS 
 		if (client->graph_wait_fd >= 0 &&
 		    (client->pollfd[WAIT_POLL_INDEX].revents & ~POLLIN)) {
 			
-			DEBUG ("\n\n\n\n\n\n\n\nWAITFD ERROR, ZOMBIE\n\n\n\n\n");
+			DEBUG ("\n\n\n\n\n\n\n\nWAITFD ERROR,"
+			       " ZOMBIE\n\n\n\n\n");
 			
 			/* our upstream "wait" connection
 			   closed, which either means that
@@ -997,10 +1192,12 @@ jack_client_thread (void *arg)
 				 * again until we get a
 				 * GraphReordered event.
 				 */
+				 
 				client->graph_wait_fd = -1;
 				client->pollmax = 1;
 			}
 		}
+	#endif
 
 		if (client->control->dead) {
 			goto zombie;
@@ -1058,7 +1255,6 @@ jack_client_thread (void *arg)
 
 			case BufferSizeChange:
 				jack_client_invalidate_port_buffers (client);
-
 				if (control->bufsize) {
 					status = control->bufsize
 						(control->nframes,
@@ -1105,7 +1301,9 @@ jack_client_thread (void *arg)
 				break;
 			}
 		}
+		
 
+#ifndef JACK_USE_MACH_THREADS 
 		if (client->pollfd[WAIT_POLL_INDEX].revents & POLLIN) {
 
 #ifdef WITH_TIMESTAMPS
@@ -1124,6 +1322,9 @@ jack_client_thread (void *arg)
 
 			control->state = Running;
 
+			/* begin preemption checking */
+			CHECK_PREEMPTION (client->engine, TRUE);
+			
 			if (control->sync_cb)
 				jack_call_sync_client (client);
 
@@ -1140,11 +1341,16 @@ jack_client_thread (void *arg)
 			if (control->timebase_cb)
 				jack_call_timebase_master (client);
 
+			/* end preemption checking */
+			CHECK_PREEMPTION (client->engine, FALSE);
+
 			control->finished_at = jack_get_microseconds();
 
 #ifdef WITH_TIMESTAMPS
 			jack_timestamp ("finished");
 #endif
+		
+#ifndef JACK_USE_MACH_THREADS 			
 			/* pass the execution token along */
 
 			DEBUG ("client finished processing at %" PRIu64
@@ -1166,12 +1372,14 @@ jack_client_thread (void *arg)
 			DEBUG ("client sent message to next stage by %" PRIu64
 			       ", client reading on graph_wait_fd==%d", 
 			       jack_get_microseconds(), client->graph_wait_fd);
+#endif
 
 #ifdef WITH_TIMESTAMPS
 			jack_timestamp ("read pending byte from wait");
 #endif
 			DEBUG("reading cleanup byte from pipe\n");
 
+#ifndef JACK_USE_MACH_THREADS
 			if ((read (client->graph_wait_fd, &c, sizeof (c))
 			     != sizeof (c))) {
 				DEBUG ("WARNING: READ FAILED!");
@@ -1183,6 +1391,7 @@ jack_client_thread (void *arg)
 				break;
 #endif
 			}
+#endif
 
 			/* check if we were killed during the process
 			 * cycle (or whatever) */
@@ -1198,6 +1407,8 @@ jack_client_thread (void *arg)
 #endif			
 
 		}
+#endif
+
 	}
 	
 	return (void *) ((intptr_t)err);
@@ -1207,7 +1418,7 @@ jack_client_thread (void *arg)
 		jack_error ("zombified - calling shutdown handler");
 		client->on_shutdown (client->on_shutdown_arg);
 	} else {
-		jack_error ("zombified - exiting from JACK");
+		jack_error ("jack_client_thread zombified - exiting from JACK");
 		jack_client_close (client);
 		/* Need a fix : possibly make client crash if
 		 * zombified without shutdown handler 
@@ -1219,7 +1430,6 @@ jack_client_thread (void *arg)
 	return 0;
 }
 
-
 #ifdef JACK_USE_MACH_THREADS
 /* real-time thread : separated from the normal client thread, it will
  * communicate with the server using fast mach RPC mechanism */
@@ -1231,6 +1441,12 @@ jack_client_process_thread (void *arg)
 	jack_client_control_t *control = client->control;
 	int err = 0;
       
+	if (client->control->thread_init) {
+	/* this means that the init callback will be called twice -taybin*/
+		DEBUG ("calling client thread init callback");
+		client->control->thread_init (client->control->thread_init_arg);
+	}
+
 	client->control->pid = getpid();
 	DEBUG ("client process thread is now running");
         
@@ -1238,50 +1454,53 @@ jack_client_process_thread (void *arg)
             
    	while (err == 0) {
 	
-                if (jack_client_suspend(client) < 0) {
-                        jack_error ("jack_client_process_thread : resume error");
-                        goto zombie;
-                }
-                
-                control->awake_at = jack_get_microseconds();
-                
-                DEBUG ("client resumed");
-                 
-                control->state = Running;
+		if (jack_client_suspend(client) < 0) {
+				jack_error ("jack_client_process_thread :resume error");
+				goto zombie;
+		}
+		
+		control->awake_at = jack_get_microseconds();
+		
+		DEBUG ("client resumed");
+		 
+		control->state = Running;
 
 		if (control->sync_cb)
 			jack_call_sync_client (client);
 
-                if (control->process) {
-                        if (control->process (control->nframes,
-					      control->process_arg) == 0) {
-                                control->state = Finished;
-                        }
-                } else {
-                        control->state = Finished;
-                }
+		if (control->process) {
+			if (control->process (control->nframes,
+				control->process_arg) == 0) {
+					control->state = Finished;
+			}
+		} else {
+			control->state = Finished;
+		}
 
 		if (control->timebase_cb)
 			jack_call_timebase_master (client);
                 
-                control->finished_at = jack_get_microseconds();
+		control->finished_at = jack_get_microseconds();
                 
 #ifdef WITH_TIMESTAMPS
-                jack_timestamp ("finished");
+		jack_timestamp ("finished");
 #endif
-                DEBUG ("client finished processing at %Lu (elapsed = %f usecs)", 
-                               control->finished_at, ((float)(control->finished_at - control->awake_at)));
+		DEBUG ("client finished processing at %Lu (elapsed = %f usecs)",
+			control->finished_at,
+			((float)(control->finished_at - control->awake_at)));
                   
-                /* check if we were killed during the process cycle (or whatever) */
+		/* check if we were killed during the process cycle
+		 * (or whatever) */
 
-                if (client->control->dead) {
-                        jack_error ("jack_client_process_thread : client->control->dead");
-                        goto zombie;
-                }
+		if (client->control->dead) {
+				jack_error ("jack_client_process_thread: "
+						"client->control->dead");
+				goto zombie;
+		}
 
-                DEBUG("process cycle fully complete\n");
+		DEBUG("process cycle fully complete\n");
 
-        }
+	}
         
  	return (void *) ((intptr_t)err);
 
@@ -1295,8 +1514,10 @@ jack_client_process_thread (void *arg)
 		jack_error ("zombified - calling shutdown handler");
 		client->on_shutdown (client->on_shutdown_arg);
 	} else {
-		jack_error ("zombified - exiting from JACK");
-		jack_client_close (client); /* Need a fix : possibly make client crash if zombified without shutdown handler */
+		jack_error ("jack_client_process_thread zombified - exiting from JACK");
+		/* Need a fix : possibly make client crash if
+		 * zombified without shutdown handler */
+		jack_client_close (client); 
 	}
 
 	pthread_exit (0);
@@ -1308,14 +1529,13 @@ jack_client_process_thread (void *arg)
 static int
 jack_start_thread (jack_client_t *client)
 {
-
  	if (client->engine->real_time) {
 
 #ifdef USE_MLOCK
-		 if (client->engine->do_mlock
-		     && (mlockall (MCL_CURRENT | MCL_FUTURE) != 0)) {
-			 jack_error ("cannot lock down memory for RT thread (%s)",
-				     strerror (errno));
+		if (client->engine->do_mlock
+		    && (mlockall (MCL_CURRENT | MCL_FUTURE) != 0)) {
+			jack_error ("cannot lock down memory for RT thread "
+				    "(%s)", strerror (errno));
 			 
 #ifdef ENSURE_MLOCK
 			 return -1;
@@ -1330,16 +1550,18 @@ jack_start_thread (jack_client_t *client)
 
 #ifdef JACK_USE_MACH_THREADS
 /* Stephane Letz : letz@grame.fr
-On MacOSX, the normal thread does not need to be real-time.
+   On MacOSX, the normal thread does not need to be real-time.
 */
-	if (jack_create_thread (&client->thread,
-				client->engine->client_priority,
-				0,
-				jack_client_thread, client)) {
+	if (jack_client_create_thread (client, 
+				       &client->thread,
+				       client->engine->client_priority,
+				       FALSE,
+				       jack_client_thread, client)) {
 		return -1;
 	}
 #else
-	if (jack_create_thread (&client->thread,
+	if (jack_client_create_thread (client,
+				&client->thread,
 				client->engine->client_priority,
 				client->engine->real_time,
 				jack_client_thread, client)) {
@@ -1347,7 +1569,7 @@ On MacOSX, the normal thread does not need to be real-time.
 	}
 
 #endif
-	
+
 #ifdef JACK_USE_MACH_THREADS
 
 	/* a secondary thread that runs the process callback and uses
@@ -1360,10 +1582,11 @@ On MacOSX, the normal thread does not need to be real-time.
 
 	*/
 
-	if (jack_create_thread(&client->process_thread,
-				client->engine->client_priority,
-				client->engine->real_time,
-				jack_client_process_thread, client)) {
+	if (jack_client_create_thread(client,
+				      &client->process_thread,
+				      client->engine->client_priority,
+				      client->engine->real_time,
+				      jack_client_process_thread, client)) {
 		return -1;
 	}
 #endif /* JACK_USE_MACH_THREADS */
@@ -1408,7 +1631,8 @@ jack_activate (jack_client_t *client)
 
 		req.type = SetClientCapabilities;
 		req.x.client_id = client->control->id;
-		
+		req.x.cap_pid = client->control->pid;
+
 		jack_client_deliver_request (client, &req);
 
 		if (req.status) {
@@ -1417,17 +1641,14 @@ jack_activate (jack_client_t *client)
 			   is using capabilities and has them
 			   (otherwise we would not get an error
 			   return) but for some reason it could not
-			   give the client the required capabilities,
-			   so for now downgrade the client so that it
-			   still runs, albeit non-realtime - nando
+			   give the client the required capabilities.
+			   For now, leave the client so that it
+			   still runs, albeit non-realtime.
 			*/
-
+			
 			jack_error ("could not receive realtime capabilities, "
 				    "client will run non-realtime");
-			/* XXX wrong, this is a property of the engine
-			client->engine->real_time = 0;
-			*/
-		}
+		} 
 	}
 #endif /* USE_CAPABILITIES */
 
@@ -1437,7 +1658,7 @@ jack_activate (jack_client_t *client)
 		pthread_cond_init (&client_ready, NULL);
 		
 		pthread_mutex_lock (&client_lock);
-		
+
 		if (jack_start_thread (client)) {
 			pthread_mutex_unlock (&client_lock);
 			return -1;
@@ -1445,7 +1666,7 @@ jack_activate (jack_client_t *client)
 
 		pthread_cond_wait (&client_ready, &client_lock);
 		pthread_mutex_unlock (&client_lock);
-		
+
 		if (!client->thread_ok) {
 			jack_error ("could not start client thread");
 			return -1;
@@ -1464,14 +1685,19 @@ jack_activate (jack_client_t *client)
 
 int 
 jack_deactivate (jack_client_t *client)
-
 {
 	jack_request_t req;
-
-	req.type = DeactivateClient;
-	req.x.client_id = client->control->id;
-
-	return jack_client_deliver_request (client, &req);
+	int rc = ESRCH;			/* already shut down */
+	
+	if (client && client->control) { /* not shut down? */
+		rc = 0;
+		if (client->control->active) { /* still active? */
+			req.type = DeactivateClient;
+			req.x.client_id = client->control->id;
+			rc = jack_client_deliver_request (client, &req);
+		}
+	}
+	return rc;
 }
 
 int
@@ -1479,9 +1705,11 @@ jack_client_close (jack_client_t *client)
 {
 	JSList *node;
 	void *status;
-	
-	if (client->control->active) {
-		jack_deactivate (client);
+	int rc;
+
+	rc = jack_deactivate (client);
+	if (rc == ESRCH) {		/* already shut down? */
+		return rc;
 	}
 
 	if (client->control->type == ClientExternal) {
@@ -1523,6 +1751,7 @@ jack_client_close (jack_client_t *client)
 			client->port_segment = NULL;
 		}
 
+#ifndef JACK_USE_MACH_THREADS
 		if (client->graph_wait_fd) {
 			close (client->graph_wait_fd);
 		}
@@ -1530,18 +1759,29 @@ jack_client_close (jack_client_t *client)
 		if (client->graph_next_fd) {
 			close (client->graph_next_fd);
 		}
+#endif		
 		
 		close (client->event_fd);
+
+		if (shutdown (client->request_fd, SHUT_RDWR)) {
+			jack_error ("could not shutdown client request socket");
+		}
+
 		close (client->request_fd);
+
 	}
 
 	for (node = client->ports; node; node = jack_slist_next (node)) {
 		free (node->data);
 	}
 	jack_slist_free (client->ports);
+	for (node = client->ports_ext; node; node = jack_slist_next (node)) {
+		free (node->data);
+	}
+	jack_slist_free (client->ports_ext);
 	jack_client_free (client);
 
-	return 0;
+	return rc;
 }	
 
 int 
@@ -1632,7 +1872,6 @@ jack_set_error_function (void (*func) (const char *))
 	jack_error_callback = func;
 }
 
-
 int 
 jack_set_graph_order_callback (jack_client_t *client,
 			       JackGraphOrderCallback callback, void *arg)
@@ -1823,6 +2062,24 @@ jack_cpu_load (jack_client_t *client)
 	return client->engine->cpu_load;
 }
 
+float
+jack_get_xrun_delayed_usecs (jack_client_t *client)
+{
+	return client->engine->xrun_delayed_usecs;
+}
+
+float
+jack_get_max_delayed_usecs (jack_client_t *client)
+{
+	return client->engine->max_delayed_usecs;
+}
+
+void
+jack_reset_max_delayed_usecs (jack_client_t *client)
+{
+	client->engine->max_delayed_usecs =  0.0f;
+}
+
 pthread_t
 jack_client_thread_id (jack_client_t *client)
 {
